from collections import OrderedDict
from typing import List

import django_tables2 as tables
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Row, LayoutObject
from crud_views.lib.crispy import CrispyModelForm, Column4, CrispyModelViewMixin, CrispyDeleteForm, Column8
from crud_views.lib.table import Table, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
    ListViewPermissionRequired,
)
from crud_views.lib.viewset import ViewSet
from django.forms.models import inlineformset_factory
from django.utils.translation import gettext as _

from app.models.questions import Question, QuestionChoice, QuestionTag, QuestionChoiceTag, QuestionChoiceTagAnnotation
from crud_views.lib.formsets import FormSets, FormSet, FormSetMixin, InlineFormSet, Formsets

cv_question = ViewSet(model=Question, name="question", icon_header="fa-solid fa-question")

EXTRA = 1


class QuestionForm(CrispyModelForm):
    class Meta:
        model = Question
        fields = ["question"]

    def get_layout_fields(self):
        return [Row(Column4("question")), Formsets()]

    @property
    def helper(self) -> FormHelper:
        # todo: is this needed?
        h = super().helper
        h.form_tag = False
        return h


class QuestionTable(Table):
    id = LinkDetailColumn()
    question = tables.Column()


class QuestionListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Question
    table_class = QuestionTable
    cv_viewset = cv_question


class QuestionDetailView(DetailViewPermissionRequired):
    model = Question
    cv_viewset = cv_question

    cv_property_display = [
        {
            "title": _("Properties"),
            "icon": "circle-question",
            "description": _("Question details"),
            "properties": [
                "id",
                {"path": "question", "detail": _("The question text")},
            ],
        },
    ]


class QuestionChoiceForm(CrispyModelForm):
    class Meta:
        model = QuestionChoice
        fields = [
            "choice",
            "help_text",
        ]


class ChoiceInlineFormSet(InlineFormSet):
    def get_helper_layout_fields(self) -> List[LayoutObject]:
        return [Row(Column4("choice"), Column4("help_text"), self.form_control_col4)]


ChoiceFormSet = inlineformset_factory(
    Question,
    QuestionChoice,
    formset=ChoiceInlineFormSet,
    form=QuestionChoiceForm,
    fields=[
        "choice",
        "help_text",  # "order",
    ],
    extra=EXTRA,
    # this adds fields to the form
    can_delete=True,
    can_delete_extra=True,
    can_order=True,
)


class QuestionTagForm(CrispyModelForm):
    class Meta:
        model = QuestionTag
        fields = ["tag"]


class QuestionTagInlineFormSet(InlineFormSet):
    def get_helper_layout_fields(self) -> List[LayoutObject]:
        return [Row(Column8("tag"), self.form_control_col4)]


QuestionTagFormSet = inlineformset_factory(
    Question,
    QuestionTag,
    formset=QuestionTagInlineFormSet,
    form=QuestionTagForm,
    fields=["tag"],
    extra=EXTRA,
    # this adds fields to the form
    can_delete=True,
    can_delete_extra=True,
    can_order=False,
)


class QuestionChoiceTagForm(CrispyModelForm):
    class Meta:
        model = QuestionChoiceTag
        fields = [
            "tag",
        ]


class QuestionChoiceTagInlineFormSet(InlineFormSet):
    def get_helper_layout_fields(self) -> List[LayoutObject]:
        return [Row(Column8("tag"), self.form_control_col4)]


QuestionChoiceTagFormSet = inlineformset_factory(
    QuestionChoice,
    QuestionChoiceTag,
    formset=QuestionChoiceTagInlineFormSet,
    form=QuestionChoiceTagForm,
    fields=["tag"],
    extra=2,  # EXTRA,
    # this adds fields to the form
    can_delete=True,
    can_delete_extra=True,
    can_order=True,
)


class QuestionChoiceTagAnnotationInlineFormSet(InlineFormSet):
    def get_helper_layout_fields(self) -> List[LayoutObject]:
        return [Row(Column8("annotation"), self.form_control_col4)]


QuestionChoiceTagAnnotationFormSet = inlineformset_factory(
    QuestionChoiceTag,
    QuestionChoiceTagAnnotation,
    formset=QuestionChoiceTagAnnotationInlineFormSet,
    form=QuestionChoiceTagForm,
    fields=["annotation"],
    extra=EXTRA,
    # this adds fields to the form
    can_delete=True,
    can_delete_extra=True,
    can_order=True,
)

cv_formsets: FormSets = FormSets(
    formsets=OrderedDict(
        choices=FormSet(  # noqa
            title=_("Choices"),
            klass=ChoiceFormSet,
            fields=["choice", "help_text"],
            pk_field="id",
            children=OrderedDict(
                tags=FormSet(
                    title=_("Tags"),
                    klass=QuestionChoiceTagFormSet,
                    fields=["tag"],
                    pk_field="id",
                    children=OrderedDict(
                        annotations=FormSet(
                            title=_("Annotations"),
                            klass=QuestionChoiceTagAnnotationFormSet,
                            fields=["annotation"],
                            pk_field="id",
                        )
                    ),
                )
            ),
        ),
        tags=FormSet(
            title=_("Tags"),
            klass=QuestionTagFormSet,
            fields=["tag"],
            pk_field="id",
        ),
    )
)


class QuestionUpdateView(CrispyModelViewMixin, FormSetMixin, UpdateViewPermissionRequired):
    model = Question
    form_class = QuestionForm
    cv_viewset = cv_question
    cv_formsets: FormSets = cv_formsets


class QuestionCreateView(CrispyModelViewMixin, FormSetMixin, CreateViewPermissionRequired):
    model = Question
    form_class = QuestionForm
    cv_viewset = cv_question
    cv_formsets: FormSets = cv_formsets


class QuestionDeleteView(CrispyModelViewMixin, DeleteViewPermissionRequired):
    model = Question
    form_class = CrispyDeleteForm
    cv_viewset = cv_question



import pytest
from django.contrib.contenttypes.models import ContentType
from django.test.client import Client

from crud_views.lib.viewset import ViewSet
from tests.lib.helper.boostrap5 import Table
from tests.test1.app.models import Vehicle, Car, Truck


@pytest.mark.django_db
def test_viewset_has_polymorphic_views(cv_vehicle: ViewSet):
    """Polymorphic viewset should have all expected views registered."""
    for view_name in ["list", "detail", "create", "create_select", "update", "delete"]:
        assert cv_vehicle.has_view(view_name), f"ViewSet missing view: {view_name}"


@pytest.mark.django_db
def test_viewset_url_patterns(cv_vehicle: ViewSet):
    """URL patterns should include polymorphic-specific routes."""
    routes = {p.name for p in cv_vehicle.urlpatterns}
    assert "vehicle-create_select" in routes
    assert "vehicle-create" in routes

    # Create URL should contain polymorphic_ctype_id
    create_patterns = [p for p in cv_vehicle.urlpatterns if p.name == "vehicle-create"]
    assert len(create_patterns) == 1
    assert "polymorphic_ctype_id" in create_patterns[0].pattern.regex.pattern


@pytest.mark.django_db
def test_list_shows_all_subtypes(client_user_vehicle_view: Client, cv_vehicle, car_sedan, truck_semi):
    """List view should show all polymorphic subtypes."""
    response = client_user_vehicle_view.get("/vehicle/")
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 2
    names = {row.columns[1].text for row in table.rows}
    assert "Sedan" in names
    assert "Semi" in names


@pytest.mark.django_db
def test_detail_car(client_user_vehicle_view: Client, cv_vehicle, car_sedan):
    """Detail view should resolve to the correct polymorphic subtype."""
    response = client_user_vehicle_view.get(f"/vehicle/{car_sedan.pk}/detail/")
    assert response.status_code == 200

    content = response.content.decode("utf-8")
    assert "Sedan" in content


@pytest.mark.django_db
def test_detail_truck(client_user_vehicle_view: Client, cv_vehicle, truck_semi):
    response = client_user_vehicle_view.get(f"/vehicle/{truck_semi.pk}/detail/")
    assert response.status_code == 200

    content = response.content.decode("utf-8")
    assert "Semi" in content


@pytest.mark.django_db
def test_create_select_shows_form(client_user_vehicle_add: Client, cv_vehicle):
    """Create-select view should show a form with polymorphic type choices."""
    response = client_user_vehicle_add.get("/vehicle/create/select/")
    assert response.status_code == 200

    content = response.content.decode("utf-8")
    assert "polymorphic_ctype_id" in content


@pytest.mark.django_db
def test_create_select_redirects_to_create(client_user_vehicle_add: Client, cv_vehicle):
    """Submitting create-select should redirect to the polymorphic create view."""
    car_ct = ContentType.objects.get_for_model(Car)
    response = client_user_vehicle_add.post("/vehicle/create/select/", {"polymorphic_ctype_id": car_ct.id})
    assert response.status_code == 302
    assert f"/vehicle/create//ct/{car_ct.id}/" in response.url


@pytest.mark.django_db
def test_create_car(client_user_vehicle_add: Client, cv_vehicle):
    """Creating a Car through the polymorphic create view."""
    car_ct = ContentType.objects.get_for_model(Car)

    # GET the form
    response = client_user_vehicle_add.get(f"/vehicle/create//ct/{car_ct.id}/")
    assert response.status_code == 200

    # POST the form
    response = client_user_vehicle_add.post(f"/vehicle/create//ct/{car_ct.id}/", {"name": "Coupe", "doors": 2})
    assert response.status_code == 302

    car = Car.objects.get(name="Coupe")
    assert car.doors == 2
    # Should also be queryable as Vehicle
    assert Vehicle.objects.filter(pk=car.pk).exists()


@pytest.mark.django_db
def test_create_truck(client_user_vehicle_add: Client, cv_vehicle):
    """Creating a Truck through the polymorphic create view."""
    truck_ct = ContentType.objects.get_for_model(Truck)

    response = client_user_vehicle_add.post(
        f"/vehicle/create//ct/{truck_ct.id}/", {"name": "Pickup", "payload_tons": 5}
    )
    assert response.status_code == 302

    truck = Truck.objects.get(name="Pickup")
    assert truck.payload_tons == 5


@pytest.mark.django_db
def test_update_car(client_user_vehicle_change: Client, cv_vehicle, car_sedan):
    """Updating a Car through the polymorphic update view."""
    response = client_user_vehicle_change.get(f"/vehicle/{car_sedan.pk}/update/")
    assert response.status_code == 200

    response = client_user_vehicle_change.post(f"/vehicle/{car_sedan.pk}/update/", {"name": "Sports Car", "doors": 2})
    assert response.status_code == 302

    car_sedan.refresh_from_db()
    assert car_sedan.name == "Sports Car"
    assert car_sedan.doors == 2


@pytest.mark.django_db
def test_update_truck(client_user_vehicle_change: Client, cv_vehicle, truck_semi):
    """Updating a Truck through the polymorphic update view."""
    response = client_user_vehicle_change.post(
        f"/vehicle/{truck_semi.pk}/update/", {"name": "Big Rig", "payload_tons": 30}
    )
    assert response.status_code == 302

    truck_semi.refresh_from_db()
    assert truck_semi.name == "Big Rig"
    assert truck_semi.payload_tons == 30


@pytest.mark.django_db
def test_delete_vehicle(client_user_vehicle_delete: Client, cv_vehicle, car_sedan):
    """Deleting a polymorphic vehicle."""
    pk = car_sedan.pk

    response = client_user_vehicle_delete.get(f"/vehicle/{pk}/delete/")
    assert response.status_code == 200

    response = client_user_vehicle_delete.post(f"/vehicle/{pk}/delete/", {"confirm": True})
    assert response.status_code == 302

    assert not Vehicle.objects.filter(pk=pk).exists()
    assert not Car.objects.filter(pk=pk).exists()


@pytest.mark.django_db
def test_polymorphic_queryset_returns_subtypes(cv_vehicle, car_sedan, truck_semi):
    """Vehicle.objects.all() should return Car and Truck instances."""
    vehicles = list(Vehicle.objects.all())
    types = {type(v) for v in vehicles}
    assert Car in types
    assert Truck in types



from django.contrib import admin

from app.models import Author


@admin.register(Author)
class AuthorAdmin(admin.ModelAdmin):
    list_display = ["first_name", "last_name", "pseudonym", "created_dt", "modified_dt"]



from django.views import generic

from crud_views.lib.settings import crud_views_settings
from crud_views.lib.view import CrudView, CrudViewPermissionRequiredMixin
from .mixins import CrudViewProcessFormMixin


class UpdateView(CrudViewProcessFormMixin, CrudView, generic.UpdateView):
    template_name = "crud_views/view_update.html"

    cv_key = "update"
    cv_path = "update"
    cv_success_key = "list"
    cv_context_actions = crud_views_settings.update_context_actions

    # texts and labels
    cv_header_template: str | None = "crud_views/snippets/header/update.html"
    cv_paragraph_template: str | None = "crud_views/snippets/paragraph/update.html"
    cv_action_label_template: str | None = "crud_views/snippets/action/update.html"
    cv_action_short_label_template: str | None = "crud_views/snippets/action_short/update.html"

    # icons
    cv_icon_action = "fa-regular fa-pen-to-square"

    # messages
    cv_message_template: str | None = "crud_views/snippets/message/update.html"

    def cv_form_valid(self, context: dict):
        """
        Handle valid form, save (update) the from instance
        """
        self.object = context["form"].save()


class UpdateViewPermissionRequired(CrudViewPermissionRequiredMixin, UpdateView):  # this file
    cv_permission = "change"



from django.utils.translation import gettext as _
from django.views import generic

from crud_views.lib.view import CrudView, CrudViewPermissionRequiredMixin
from crud_views.lib.settings import crud_views_settings

# crispy may not be installed
try:
    from crispy_forms.helper import FormHelper
    from crispy_forms import layout
except ImportError:
    FormHelper = object
    layout = None


class ListView(CrudView, generic.ListView):
    template_name = "crud_views/view_list.html"

    cv_pk: bool = False  # does not need primary key
    cv_key = "list"
    cv_path = ""
    cv_object = False
    cv_list_actions = crud_views_settings.list_actions
    cv_context_actions = crud_views_settings.list_context_actions

    # texts and labels
    cv_header_template: str | None = "crud_views/snippets/header/list.html"
    cv_paragraph_template: str | None = "crud_views/snippets/paragraph/list.html"
    cv_action_label_template: str | None = "crud_views/snippets/action/list.html"
    cv_action_short_label_template: str | None = "crud_views/snippets/action_short/list.html"
    cv_filter_header_template: str | None = "crud_views/snippets/header/filter.html"
    cv_filter_header_template_code: str | None = None

    # icons
    cv_icon_action = "fa-regular fa-rectangle-list"

    # todo: add check for cv_filter_header_template/cv_filter_header_template_code

    @staticmethod
    def cv_get_filter_icon() -> str:
        """
        Currently there is only one global filter icon
        """
        return crud_views_settings.filter_icon

    @property
    def cv_filter_header(self) -> str:
        """
        Get the filter header label
        """
        return self.render_snippet(
            self.cv_get_meta(),
            self.cv_filter_header_template,
            self.cv_filter_header_template_code,
        )


class ListViewPermissionRequired(CrudViewPermissionRequiredMixin, ListView):
    cv_permission = "view"


class ListViewFilterFormHelper(FormHelper):
    """
    Form helper for the filter form
    """

    form_method = "GET"  # filter parameters are always GET
    form_tag = False  # todo really?, just add hidden stuff

    def __init__(self, request, form=None):
        super().__init__(form)

        # add filter control buttons
        self.add_input(
            layout.Submit("submit", "Apply Filter", css_id="filter-button"),
        )
        self.add_input(
            layout.Reset(
                "reset",
                _("Reset Filter"),
                css_id="filter-button-reset",
                css_class=crud_views_settings.filter_reset_button_css_class,
            )
        )

        # add hidden fields with control values
        sort = request.GET.get("sort") or ""
        self.add_input(
            layout.Hidden("sort", sort),
        )



import pytest
from django.test.client import Client


@pytest.mark.django_db
def test_anonymous_redirected(client: Client, cv_author, author_douglas_adams):
    pk = author_douglas_adams.pk

    urls = [
        "/author/",
        f"/author/{pk}/detail/",
        "/author/create/",
        f"/author/{pk}/update/",
        f"/author/{pk}/delete/",
    ]
    for url in urls:
        response = client.get(url)
        assert response.status_code == 302, f"Expected 302 for anonymous on {url}, got {response.status_code}"


@pytest.mark.django_db
def test_no_permission_denied(client_user_a: Client, cv_author, author_douglas_adams):
    pk = author_douglas_adams.pk

    urls = [
        "/author/",
        f"/author/{pk}/detail/",
        "/author/create/",
        f"/author/{pk}/update/",
        f"/author/{pk}/delete/",
    ]
    for url in urls:
        response = client_user_a.get(url)
        assert response.status_code == 403, f"Expected 403 for no-perm user on {url}, got {response.status_code}"


@pytest.mark.django_db
def test_add_permission_can_create_only(client_user_author_add: Client, cv_author, author_douglas_adams):
    pk = author_douglas_adams.pk

    # Can access create
    response = client_user_author_add.get("/author/create/")
    assert response.status_code == 200

    # Cannot access update or delete
    response = client_user_author_add.get(f"/author/{pk}/update/")
    assert response.status_code == 403

    response = client_user_author_add.get(f"/author/{pk}/delete/")
    assert response.status_code == 403


@pytest.mark.django_db
def test_change_permission_can_update_only(client_user_author_change: Client, cv_author, author_douglas_adams):
    pk = author_douglas_adams.pk

    # Can access update
    response = client_user_author_change.get(f"/author/{pk}/update/")
    assert response.status_code == 200

    # Cannot access create or delete
    response = client_user_author_change.get("/author/create/")
    assert response.status_code == 403

    response = client_user_author_change.get(f"/author/{pk}/delete/")
    assert response.status_code == 403


@pytest.mark.django_db
def test_delete_permission_can_delete_only(client_user_author_delete: Client, cv_author, author_douglas_adams):
    pk = author_douglas_adams.pk

    # Can access delete
    response = client_user_author_delete.get(f"/author/{pk}/delete/")
    assert response.status_code == 200

    # Cannot access create or update
    response = client_user_author_delete.get("/author/create/")
    assert response.status_code == 403

    response = client_user_author_delete.get(f"/author/{pk}/update/")
    assert response.status_code == 403



import django_tables2 as tables
from crispy_forms.layout import Row

from app.models import Book
from crud_views.lib.crispy import CrispyModelForm, Column4, Column2, CrispyModelViewMixin, CrispyDeleteForm
from crud_views.lib.table import Table, UUIDLinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
    ListViewPermissionRequired,
    CreateViewParentMixin,
    MessageMixin,
)
from crud_views.lib.viewset import ViewSet, ParentViewSet

cv_book = ViewSet(
    model=Book,
    name="book",
    parent=ParentViewSet(name="author"),
    icon_header="fa-regular fa-address-book",  # <i class="fa-regular fa-address-book"></i>
)


class BookCreateForm(CrispyModelForm):
    submit_label = "Create"

    class Meta:
        model = Book
        fields = ["title", "price"]

    def get_layout_fields(self):
        return Row(Column4("title"), Column2("price"))


class BookUpdateForm(BookCreateForm):
    """
    Update form has the same fields as the create form
    """


class BookTable(Table):
    id = UUIDLinkDetailColumn()
    title = tables.Column()
    price = tables.Column()
    author = tables.Column()


class BookListView(ListViewTableMixin, ListViewPermissionRequired):
    cv_viewset = cv_book
    # cv_list_actions = ["detail", "update", "delete"]

    table_class = BookTable


class BookDetailView(DetailViewPermissionRequired):
    cv_viewset = cv_book

    cv_property_display = [
        {
            "title": "Attributes",
            "icon": "book-open",
            "description": "Book details and metadata",
            "properties": [
                "id",
                "title",
                {"path": "price", "detail": "Retail price in EUR"},
                "author",
                "created_dt",
                "modified_dt",
            ],
        },
    ]


class BookUpdateView(CrispyModelViewMixin, MessageMixin, UpdateViewPermissionRequired):
    form_class = BookUpdateForm
    cv_viewset = cv_book


class BookCreateView(CrispyModelViewMixin, MessageMixin, CreateViewParentMixin, CreateViewPermissionRequired):
    form_class = BookCreateForm
    cv_viewset = cv_book


class BookDeleteView(CrispyModelViewMixin, MessageMixin, DeleteViewPermissionRequired):
    form_class = CrispyDeleteForm
    cv_viewset = cv_book



from __future__ import annotations

import json
from typing import Any, List, Iterable, Tuple

from django.forms.forms import BaseForm
from django.forms.formsets import BaseFormSet
from pydantic import BaseModel, Field

try:
    from polymorphic.formsets import BasePolymorphicInlineFormSet
except ImportError:
    BasePolymorphicInlineFormSet = None


class XForm(BaseModel, arbitrary_types_allowed=True):
    level: int
    prefix_key: str
    form: BaseForm
    parent: XFormSet | None = None
    formsets: List[XFormSet] = Field(default_factory=lambda: list())

    def __str__(self):
        return f"XForm(prefix={self.prefix},prefix_key={self.prefix_key},key={self.key},level={self.level})"

    __repr__ = __str__

    @property
    def key(self) -> str:
        """return the formset key from the parent formset"""
        return self.parent.key

    @property
    def prefix(self) -> str:
        """return the form prefix"""
        return self.form.prefix

    @property
    def data(self) -> dict:
        return dict(
            key=self.key,
            prefix=self.prefix,
            prefix_key=self.prefix_key,
            formset_prefix=self.parent.prefix,
            pk=str(self.parent.pk),
        )

    @property
    def json_data(self) -> str:
        return json.dumps(self.data)  # .replace('"', "'")

    def is_valid(self) -> Iterable[Tuple[Any, bool]]:
        for x_formset in self.formsets:
            yield from x_formset.is_valid()

    def save(self, commit: bool = True, delete: bool = False):
        for x_formset in self.formsets:
            x_formset.save(commit=commit, delete=delete)

    @property
    def helper(self):
        # in polymorphic formsets we need to get the helper from the form
        klass = self.parent.formset.klass
        if BasePolymorphicInlineFormSet and issubclass(klass, BasePolymorphicInlineFormSet):
            return self.form.helper

        # in no-polymorphic formsets we can get the helper from the formset
        return self.parent.instance.helper


class XFormSet(BaseModel, arbitrary_types_allowed=True):
    level: int
    prefix_key: str
    formset: Any
    instance: BaseFormSet
    management_form: Any
    forms: List[XForm] = Field(default_factory=lambda: list())
    parent: XForm | None = None
    start_at_rows: bool = False  # todo: bad name, improve

    def __str__(self):
        return f"XFormSet({self.prefix},{self.prefix_key},{self.level})"

    __repr__ = __str__

    @property
    def has_parent(self) -> bool:
        return self.parent is not None

    @property
    def key(self) -> str:
        """return the formset key"""
        return self.formset.key

    @property
    def pk(self) -> str:
        """return the formset key"""
        return self.formset.pk

    @property
    def prefix(self) -> str:
        """return the formset prefix KEY-SUFFIX"""
        return f"{self.key}-{self.prefix_key}"

    @property
    def parent_prefix(self) -> str:
        if self.has_parent:
            return self.parent.prefix
        return ""

    @property
    def parent_prefix_key(self) -> str:
        if self.has_parent:
            return self.parent.prefix_key
        return ""

    @property
    def hierarchy(self) -> List[str]:
        return self.formset.hierarchy

    @property
    def hierarchy_str(self) -> str:
        return "|".join(self.hierarchy)

    @property
    def title(self) -> str:
        return self.formset.title

    @property
    def has_pre_col(self) -> bool:
        return self.level > 0

    @property
    def col(self) -> int:
        return 11 if self.level > 0 else 12

    @property
    def col_pre(self) -> int:
        return 1 if self.level > 0 else 0

    @property
    def css_col(self) -> str:
        return f"col-{self.col}"

    @property
    def css_col_pre(self) -> str:
        return f"col-{self.col_pre}"

    @property
    def data(self) -> dict:
        return dict(
            key=self.key,
            prefix=self.prefix,
            prefix_key=self.prefix_key,
            hierarchy=self.hierarchy,
            parent_prefix=self.parent_prefix,
            parent_prefix_key=self.parent_prefix_key,
            can_delete=self.formset.can_delete,
            can_delete_extra=self.formset.can_delete_extra,
            can_order=self.formset.can_order,
            edit_only=self.formset.edit_only,
            path=self.formset.path,
            fields=self.formset.fields,
            pk_field=self.formset.pk_field,
            pk=self.pk.value,
        )

    @property
    def json_data(self) -> str:
        return json.dumps(self.data)

    def is_valid(self) -> Iterable[Tuple[Any, bool]]:
        yield self, self.instance.is_valid()
        for x_form in self.forms:
            yield from x_form.is_valid()

    def save(self, commit: bool = True, delete=False):
        """
        Nested save and delete
        """

        def get_x_form(form):
            for x_form in self.forms:
                if form == x_form.form:
                    return x_form
            assert Exception("not found")

        # NOTE: instance.ordered_forms DOES NOT INCLUDE DELETED FORMS ;-)
        can_order = self.formset.can_order

        # handle deleted
        delete_forms = [f for f in self.instance.forms if delete is True or f.cleaned_data.get("DELETE", False)]
        regular_forms = [f for f in self.instance.forms if f not in delete_forms]
        ordered_forms = self.instance.ordered_forms if can_order else []
        update_forms = ordered_forms if can_order else regular_forms

        # process forms to delete
        for form in delete_forms:
            # get x-form to process nested forms
            x_form = get_x_form(form)

            # delete is instance and pk
            if form.instance and form.instance.pk:
                form.instance.delete()

            # deleted nested forms
            x_form.save(commit=commit, delete=True)

        # do not continue in case of nested delete
        if delete:
            return

        for form in update_forms:
            # just make sure x-form exists for this form
            get_x_form(form)

            # check for changes
            has_changed = form.has_changed()
            if not has_changed:
                continue

            # save instance
            instance = form.save(commit=False)

            # update order
            if can_order:
                order_value = form.cleaned_data.get("ORDER")
                if order_value is not None:
                    instance.order = order_value

            # finally save instance
            instance.save()
            form.save_m2m()

        # now save the x_forms
        for x_form in self.forms:
            x_form.save(commit=commit)



"""Generate UI mockup PNGs for django-crud-views documentation."""

from PIL import Image, ImageDraw, ImageFont
import os

# --- Colour palette (Bootstrap 5 inspired) ---
WHITE = "#ffffff"
LIGHT_BG = "#f8f9fa"
DARK_TEXT = "#212529"
MUTED_TEXT = "#6c757d"
PRIMARY = "#0d6efd"
PRIMARY_HOVER = "#0b5ed7"
SECONDARY = "#6c757d"
SUCCESS = "#198754"
DANGER = "#dc3545"
WARNING = "#ffc107"
BORDER = "#dee2e6"
CARD_BORDER = "#ced4da"
TABLE_STRIPE = "#f2f2f2"
LINK_COLOR = "#0d6efd"
HEADER_BG = "#e9ecef"

W = 960
PADDING = 30
INNER_W = W - 2 * PADDING


# Font helpers
def get_fonts():
    """Try to find suitable fonts on the system."""
    font_paths = [
        "/usr/share/fonts/truetype/dejavu/DejaVuSans.ttf",
        "/usr/share/fonts/truetype/liberation/LiberationSans-Regular.ttf",
        "/usr/share/fonts/TTF/DejaVuSans.ttf",
    ]
    bold_paths = [
        "/usr/share/fonts/truetype/dejavu/DejaVuSans-Bold.ttf",
        "/usr/share/fonts/truetype/liberation/LiberationSans-Bold.ttf",
        "/usr/share/fonts/TTF/DejaVuSans-Bold.ttf",
    ]
    mono_paths = [
        "/usr/share/fonts/truetype/dejavu/DejaVuSansMono.ttf",
        "/usr/share/fonts/truetype/liberation/LiberationMono-Regular.ttf",
        "/usr/share/fonts/TTF/DejaVuSansMono.ttf",
    ]

    def find(paths, size):
        for p in paths:
            if os.path.exists(p):
                return ImageFont.truetype(p, size)
        return ImageFont.load_default()

    return {
        "title": find(bold_paths, 22),
        "heading": find(bold_paths, 17),
        "body": find(font_paths, 14),
        "body_bold": find(bold_paths, 14),
        "small": find(font_paths, 12),
        "small_bold": find(bold_paths, 12),
        "mono": find(mono_paths, 12),
        "icon": find(font_paths, 13),
    }


FONTS = get_fonts()


# --- Drawing primitives ---


def rounded_rect(draw, xy, fill=None, outline=None, radius=6, width=1):
    x0, y0, x1, y1 = xy
    draw.rounded_rectangle(xy, radius=radius, fill=fill, outline=outline, width=width)


def draw_button(draw, x, y, label, color=PRIMARY, text_color=WHITE, outlined=False, small=False):
    font = FONTS["small_bold"] if small else FONTS["body_bold"]
    tw = draw.textlength(label, font=font)
    px, py = (10, 4) if small else (16, 8)
    w = int(tw + 2 * px)
    h = int((FONTS["small"].size if small else FONTS["body"].size) + 2 * py)
    if outlined:
        rounded_rect(draw, (x, y, x + w, y + h), fill=WHITE, outline=color, radius=4, width=2)
        draw.text((x + px, y + py), label, fill=color, font=font)
    else:
        rounded_rect(draw, (x, y, x + w, y + h), fill=color, radius=4)
        draw.text((x + px, y + py), label, fill=text_color, font=font)
    return w, h


def draw_card(draw, x, y, w, h, header_text=None, header_icon=None):
    rounded_rect(draw, (x, y, x + w, y + h), fill=WHITE, outline=CARD_BORDER, radius=6)
    top = y
    if header_text:
        hh = 40
        rounded_rect(draw, (x, y, x + w, y + hh), fill=HEADER_BG, outline=CARD_BORDER, radius=6)
        # Redraw bottom corners as square to join with card body
        draw.rectangle((x + 1, y + hh - 8, x + w - 1, y + hh), fill=HEADER_BG)
        draw.line((x, y + hh, x + w, y + hh), fill=CARD_BORDER)
        text = header_text
        if header_icon:
            text = f"{header_icon}  {header_text}"
        draw.text((x + 14, y + 11), text, fill=DARK_TEXT, font=FONTS["body_bold"])
        top = y + hh
    return top


def draw_table(draw, x, y, col_widths, headers, rows, w=None):
    if w is None:
        w = sum(col_widths)
    row_h = 32
    # Header
    draw.rectangle((x, y, x + w, y + row_h), fill=HEADER_BG)
    draw.line((x, y + row_h, x + w, y + row_h), fill=BORDER)
    cx = x
    for i, hdr in enumerate(headers):
        draw.text((cx + 10, y + 8), hdr, fill=DARK_TEXT, font=FONTS["small_bold"])
        cx += col_widths[i]
    cur_y = y + row_h
    for ri, row in enumerate(rows):
        bg = TABLE_STRIPE if ri % 2 == 1 else WHITE
        draw.rectangle((x, cur_y, x + w, cur_y + row_h), fill=bg)
        draw.line((x, cur_y + row_h, x + w, cur_y + row_h), fill=BORDER)
        cx = x
        for ci, cell in enumerate(row):
            color = DARK_TEXT
            font = FONTS["small"]
            if isinstance(cell, tuple):
                cell, color, font = cell[0], cell[1], cell[2] if len(cell) > 2 else FONTS["small"]
            draw.text((cx + 10, cur_y + 8), str(cell), fill=color, font=font)
            cx += col_widths[ci]
        cur_y += row_h
    # Border
    draw.rectangle((x, y, x + w, cur_y), outline=BORDER)
    return cur_y


def draw_form_field(draw, x, y, label, value="", w=400, required=False):
    label_text = f"{label} *" if required else label
    draw.text((x, y), label_text, fill=DARK_TEXT, font=FONTS["body_bold"])
    fy = y + 22
    fh = 34
    rounded_rect(draw, (x, fy, x + w, fy + fh), fill=WHITE, outline=BORDER, radius=4)
    if value:
        draw.text((x + 10, fy + 8), value, fill=DARK_TEXT, font=FONTS["body"])
    return fy + fh + 12


def draw_breadcrumb(draw, x, y, items):
    cx = x
    for i, item in enumerate(items):
        color = MUTED_TEXT if i < len(items) - 1 else DARK_TEXT
        font = FONTS["small"] if i < len(items) - 1 else FONTS["small_bold"]
        draw.text((cx, y), item, fill=color, font=font)
        cx += draw.textlength(item, font=font) + 4
        if i < len(items) - 1:
            draw.text((cx, y), "/", fill=MUTED_TEXT, font=FONTS["small"])
            cx += draw.textlength("/", font=FONTS["small"]) + 4
    return y + 20


def draw_context_actions(draw, x, y, actions):
    """Draw a Bootstrap 5 button group for context actions."""
    cx = x
    gap = 4
    for label, color, outlined in actions:
        bw, bh = draw_button(draw, cx, y, label, color=color, outlined=outlined)
        cx += bw + gap
    return y + bh + 10


def draw_page_header(draw, y, title, breadcrumbs=None, context_actions=None):
    """Draw page header with title, breadcrumbs, and context action buttons."""
    if breadcrumbs:
        y = draw_breadcrumb(draw, PADDING, y, breadcrumbs)
        y += 4

    draw.text((PADDING, y), title, fill=DARK_TEXT, font=FONTS["title"])
    y += 32

    if context_actions:
        y = draw_context_actions(draw, PADDING, y, context_actions)
        y += 6

    # Divider
    draw.line((PADDING, y, W - PADDING, y), fill=BORDER)
    y += 16
    return y


def create_image(height):
    img = Image.new("RGB", (W, height), LIGHT_BG)
    draw = ImageDraw.Draw(img)
    return img, draw


# --- Mockup generators ---


def mockup_list_view():
    H = 580
    img, draw = create_image(H)

    y = PADDING
    y = draw_page_header(
        draw,
        y,
        "Authors",
        breadcrumbs=["Home", "Authors"],
        context_actions=[
            ("\u2190  Parent", SECONDARY, True),
            ("\u2261  Filter", PRIMARY, True),
            ("+  Create", PRIMARY, True),
        ],
    )

    # Filter card (collapsed state indicator)
    card_y = y
    card_h = 90
    body_y = draw_card(draw, PADDING, card_y, INNER_W, card_h, header_text="Filter", header_icon="\u25bc")
    # Filter fields inside card
    fx = PADDING + 14
    fy = body_y + 10
    draw.text((fx, fy), "First name", fill=MUTED_TEXT, font=FONTS["small"])
    rounded_rect(draw, (fx, fy + 16, fx + 200, fy + 40), fill=WHITE, outline=BORDER, radius=3)
    draw.text((fx + 220, fy), "Last name", fill=MUTED_TEXT, font=FONTS["small"])
    rounded_rect(draw, (fx + 220, fy + 16, fx + 420, fy + 40), fill=WHITE, outline=BORDER, radius=3)

    y = card_y + card_h + 16

    # Table
    cols = [50, 200, 200, 200, INNER_W - 650]
    headers = ["#", "First Name", "Last Name", "Email", "Actions"]
    rows = [
        ["1", "Douglas", "Adams", "d.adams@example.com", ""],
        ["2", "Terry", "Pratchett", "t.pratchett@example.com", ""],
        ["3", "Isaac", "Asimov", "i.asimov@example.com", ""],
        ["4", "Ursula", "Le Guin", "u.leguin@example.com", ""],
        ["5", "Arthur", "Clarke", "a.clarke@example.com", ""],
    ]
    end_y = draw_table(draw, PADDING, y, cols, headers, rows, w=INNER_W)

    # Draw action buttons in last column of each row
    for ri in range(len(rows)):
        row_y = y + 32 + ri * 32  # header height + row offset
        ax = PADDING + sum(cols[:-1]) + 6
        bw, _ = draw_button(draw, ax, row_y + 4, "\u2139", color=PRIMARY, outlined=True, small=True)
        bw2, _ = draw_button(draw, ax + bw + 3, row_y + 4, "\u270e", color=PRIMARY, outlined=True, small=True)
        draw_button(draw, ax + bw + bw2 + 6, row_y + 4, "\u2717", color=DANGER, outlined=True, small=True)

    # Pagination
    py = end_y + 12
    draw.text((PADDING, py), "Showing 1-5 of 12 results", fill=MUTED_TEXT, font=FONTS["small"])
    px = W - PADDING - 200
    for i, (lbl, active) in enumerate([("\u2190", False), ("1", True), ("2", False), ("3", False), ("\u2192", False)]):
        bx = px + i * 38
        if active:
            rounded_rect(draw, (bx, py - 2, bx + 32, py + 24), fill=PRIMARY, radius=4)
            draw.text((bx + 10, py + 2), lbl, fill=WHITE, font=FONTS["small_bold"])
        else:
            rounded_rect(draw, (bx, py - 2, bx + 32, py + 24), fill=WHITE, outline=BORDER, radius=4)
            draw.text((bx + 10, py + 2), lbl, fill=PRIMARY, font=FONTS["small"])

    return img


def mockup_detail_view():
    H = 460
    img, draw = create_image(H)

    y = PADDING
    y = draw_page_header(
        draw,
        y,
        "Author: Douglas Adams",
        breadcrumbs=["Home", "Authors", "Douglas Adams"],
        context_actions=[
            ("\u2190  List", SECONDARY, True),
            ("\u270e  Update", PRIMARY, True),
            ("\u2717  Delete", DANGER, True),
        ],
    )

    # Detail card - split layout
    card_x = PADDING
    card_w = INNER_W
    card_h = 300
    body_y = draw_card(draw, card_x, y, card_w, card_h, header_text="Details")

    # Property rows in a two-column split layout
    props = [
        ("First Name", "Douglas"),
        ("Last Name", "Adams"),
        ("Email", "d.adams@example.com"),
        ("Date of Birth", "March 11, 1952"),
        ("Bio", "English author, best known for The Hitchhiker's Guide..."),
        ("Active", "\u2713  Yes"),
    ]

    py = body_y + 14
    label_w = 160
    for label, value in props:
        # Label on left (muted), value on right
        draw.text((card_x + 16, py), label, fill=MUTED_TEXT, font=FONTS["body_bold"])
        val_color = SUCCESS if value.startswith("\u2713") else DARK_TEXT
        draw.text((card_x + 16 + label_w, py), value, fill=val_color, font=FONTS["body"])
        py += 30
        # Separator line
        draw.line((card_x + 14, py - 6, card_x + card_w - 14, py - 6), fill=BORDER)

    return img


def mockup_create_view():
    H = 500
    img, draw = create_image(H)

    y = PADDING
    y = draw_page_header(
        draw,
        y,
        "Create Author",
        breadcrumbs=["Home", "Authors", "Create"],
        context_actions=[
            ("\u2190  List", SECONDARY, True),
        ],
    )

    # Form card
    card_h = 320
    body_y = draw_card(draw, PADDING, y, INNER_W, card_h)
    fy = body_y + 16

    fw = INNER_W - 60
    fx = PADDING + 20
    fy = draw_form_field(draw, fx, fy, "First Name", required=True, w=fw)
    fy = draw_form_field(draw, fx, fy, "Last Name", required=True, w=fw)
    fy = draw_form_field(draw, fx, fy, "Email", w=fw)
    fy = draw_form_field(draw, fx, fy, "Bio", w=fw)

    # Buttons
    by = fy + 8
    bw, bh = draw_button(draw, fx, by, "  Save  ", color=SUCCESS)
    draw_button(draw, fx + bw + 10, by, "  Cancel  ", color=SECONDARY, outlined=True)

    return img


def mockup_update_view():
    H = 500
    img, draw = create_image(H)

    y = PADDING
    y = draw_page_header(
        draw,
        y,
        "Update Author: Douglas Adams",
        breadcrumbs=["Home", "Authors", "Douglas Adams", "Update"],
        context_actions=[
            ("\u2190  List", SECONDARY, True),
        ],
    )

    # Form card
    card_h = 320
    body_y = draw_card(draw, PADDING, y, INNER_W, card_h)
    fy = body_y + 16

    fw = INNER_W - 60
    fx = PADDING + 20
    fy = draw_form_field(draw, fx, fy, "First Name", value="Douglas", required=True, w=fw)
    fy = draw_form_field(draw, fx, fy, "Last Name", value="Adams", required=True, w=fw)
    fy = draw_form_field(draw, fx, fy, "Email", value="d.adams@example.com", w=fw)
    fy = draw_form_field(draw, fx, fy, "Bio", value="English author, best known for...", w=fw)

    # Buttons
    by = fy + 8
    bw, bh = draw_button(draw, fx, by, "  Save  ", color=SUCCESS)
    draw_button(draw, fx + bw + 10, by, "  Cancel  ", color=SECONDARY, outlined=True)

    return img


def mockup_delete_view():
    H = 340
    img, draw = create_image(H)

    y = PADDING
    y = draw_page_header(
        draw,
        y,
        "Delete Author: Douglas Adams",
        breadcrumbs=["Home", "Authors", "Douglas Adams", "Delete"],
        context_actions=[
            ("\u2190  List", SECONDARY, True),
        ],
    )

    # Warning card
    card_h = 170
    body_y = draw_card(draw, PADDING, y, INNER_W, card_h)

    # Warning message
    wy = body_y + 20
    draw.text(
        (PADDING + 20, wy), "\u26a0  Are you sure you want to delete this author?", fill=DANGER, font=FONTS["heading"]
    )
    wy += 34

    # Object summary
    draw.text((PADDING + 20, wy), "Author:", fill=MUTED_TEXT, font=FONTS["body_bold"])
    draw.text((PADDING + 90, wy), "Douglas Adams", fill=DARK_TEXT, font=FONTS["body"])
    wy += 26
    draw.text((PADDING + 20, wy), "This action cannot be undone.", fill=MUTED_TEXT, font=FONTS["small"])

    # Buttons
    by = wy + 30
    bw, bh = draw_button(draw, PADDING + 20, by, "  Confirm Delete  ", color=DANGER)
    draw_button(draw, PADDING + 20 + bw + 10, by, "  Cancel  ", color=SECONDARY, outlined=True)

    return img


def mockup_manage_view():
    H = 580
    img, draw = create_image(H)

    y = PADDING
    y = draw_page_header(
        draw,
        y,
        "Manage: Author ViewSet",
        breadcrumbs=["Home", "Authors", "Manage"],
        context_actions=[
            ("\u2190  List", SECONDARY, True),
        ],
    )

    # Properties table
    draw.text((PADDING, y), "ViewSet Properties", fill=DARK_TEXT, font=FONTS["heading"])
    y += 26
    cols = [200, INNER_W - 200]
    props = [
        ["name", "author"],
        ["prefix", "app"],
        ["model", "app.Author"],
        ["pk", "INT"],
        ["pk_name", "pk"],
    ]
    y = draw_table(draw, PADDING, y, cols, ["Property", "Value"], props, w=INNER_W)
    y += 20

    # Permissions table
    draw.text((PADDING, y), "Permissions", fill=DARK_TEXT, font=FONTS["heading"])
    y += 26
    pcols = [200, 300, INNER_W - 500]
    perms = [
        ["view", "app.view_author", ("\u2713 Granted", SUCCESS)],
        ["add", "app.add_author", ("\u2713 Granted", SUCCESS)],
        ["change", "app.change_author", ("\u2713 Granted", SUCCESS)],
        ["delete", "app.delete_author", ("\u2713 Granted", SUCCESS)],
    ]
    y = draw_table(draw, PADDING, y, pcols, ["Permission", "Django Permission", "Access"], perms, w=INNER_W)

    return img


# --- Main ---


def main():
    out_dir = os.path.join(os.path.dirname(__file__), "..", "docs", "img")
    os.makedirs(out_dir, exist_ok=True)

    mockups = {
        "view_list": mockup_list_view,
        "view_detail": mockup_detail_view,
        "view_create": mockup_create_view,
        "view_update": mockup_update_view,
        "view_delete": mockup_delete_view,
        "view_manage": mockup_manage_view,
    }

    for name, func in mockups.items():
        path = os.path.join(out_dir, f"{name}.png")
        img = func()
        img.save(path, "PNG", optimize=True)
        print(f"  {path}")

    print(f"\nGenerated {len(mockups)} mockups in {out_dir}/")


if __name__ == "__main__":
    main()



from django.http import HttpResponseRedirect
from django.views import generic
from django.views.generic.detail import SingleObjectMixin

from crud_views.lib.view import CrudView, CrudViewPermissionRequiredMixin


class ActionView(CrudView, SingleObjectMixin, generic.View):
    cv_list_action_method = "post"

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)
        result = self.action(context)
        if result:
            self.cv_action_success_hook(context)
        else:
            self.cv_action_error_hook(context)
        # todo: evaluate result
        # todo: message
        url = self.get_success_url()
        return HttpResponseRedirect(url)

    def action(self, context: dict) -> bool:
        raise NotImplementedError("Action not implemented")

    def cv_action_success_hook(self, context: dict) -> None:
        """
        Hook for additional actions after the main action is performed.
        """
        pass

    def cv_action_error_hook(self, context: dict) -> None:
        """
        Hook for additional actions after the main action is performed.
        """
        pass


class ActionViewPermissionRequired(CrudViewPermissionRequiredMixin, ActionView):  # this file
    cv_permission = "change"



# Generated by Django 5.1.4 on 2024-12-27 13:45

import django.db.models.deletion
import uuid
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Author",
            fields=[
                ("order", models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order")),
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ("first_name", models.CharField(max_length=100)),
                ("last_name", models.CharField(max_length=100)),
                ("pseudonym", models.CharField(blank=True, max_length=100, null=True)),
                ("created_dt", models.DateTimeField(auto_now_add=True)),
                ("modified_dt", models.DateTimeField(auto_now=True)),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Bar",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="Foo",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="Baz",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100)),
                ("bar", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.bar")),
            ],
        ),
        migrations.CreateModel(
            name="Book",
            fields=[
                ("order", models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order")),
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ("title", models.CharField(max_length=100)),
                ("price", models.DecimalField(decimal_places=2, max_digits=10)),
                ("created_dt", models.DateTimeField(auto_now_add=True)),
                ("modified_dt", models.DateTimeField(auto_now=True)),
                ("author", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.author")),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="bar",
            name="foo",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.foo"),
        ),
    ]



import django_tables2 as tables
from crispy_forms.layout import Row

from app.models import Foo
from crud_views.lib.crispy import CrispyModelForm, Column4, CrispyModelViewMixin, CrispyDeleteForm
from crud_views.lib.table import Table, LinkChildColumn, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
    ListViewPermissionRequired,
)
from crud_views.lib.viewset import ViewSet

cv_foo = ViewSet(model=Foo, name="foo", icon_header="fa-solid fa-paw")


class FooForm(CrispyModelForm):
    submit_label = "Create"

    class Meta:
        model = Foo
        fields = ["name"]

    def get_layout_fields(self):
        return Row(Column4("name"))


class FooTable(Table):
    id = LinkDetailColumn()
    name = tables.Column()
    bar = LinkChildColumn(name="bar", verbose_name="Bar", empty_values=())


class FooListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Foo
    table_class = FooTable
    cv_viewset = cv_foo
    cv_list_actions = ["detail", "update", "delete"]


class FooDetailView(DetailViewPermissionRequired):
    model = Foo
    cv_viewset = cv_foo

    cv_property_display = [
        {
            "title": "Properties",
            "icon": "paw",
            "description": "Foo attributes",
            "properties": [
                "id",
                "name",
            ],
        },
    ]


class FooUpdateView(CrispyModelViewMixin, UpdateViewPermissionRequired):
    model = Foo
    form_class = FooForm
    cv_viewset = cv_foo


class FooCreateView(CrispyModelViewMixin, CreateViewPermissionRequired):
    model = Foo
    form_class = FooForm
    cv_viewset = cv_foo


class FooDeleteView(CrispyModelViewMixin, DeleteViewPermissionRequired):
    model = Foo
    form_class = CrispyDeleteForm
    cv_viewset = cv_foo



import pytest

from crud_views.lib.viewset import ViewSet


@pytest.mark.django_db
def test_viewset_has_views(cv_author: ViewSet):
    expected_views = ["list", "detail", "create", "update", "delete"]
    for view_name in expected_views:
        assert cv_author.has_view(view_name), f"ViewSet missing view: {view_name}"


@pytest.mark.django_db
def test_viewset_url_patterns(cv_author: ViewSet):
    patterns = cv_author.urlpatterns
    assert len(patterns) >= 5


@pytest.mark.django_db
def test_viewset_permissions_mapping(cv_author: ViewSet):
    permissions = cv_author.permissions
    assert "add" in permissions
    assert "change" in permissions
    assert "delete" in permissions
    assert "view" in permissions

    for key, value in permissions.items():
        assert "." in value, f"Permission {key} should be in 'app.codename' format, got {value}"


@pytest.mark.django_db
def test_viewset_get_router_name(cv_author: ViewSet):
    assert cv_author.get_router_name("list") == "author-list"
    assert cv_author.get_router_name("detail") == "author-detail"
    assert cv_author.get_router_name("create") == "author-create"
    assert cv_author.get_router_name("update") == "author-update"
    assert cv_author.get_router_name("delete") == "author-delete"



import threading
from collections import OrderedDict
from functools import cached_property
from typing import Dict, List, Type, Any, Iterable, ClassVar

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models import Model, QuerySet, Q
from django.urls import re_path, URLResolver
from django.utils.translation import gettext as _
from pydantic import BaseModel, PrivateAttr, Field, model_validator
from typing_extensions import Self

from crud_views.lib.exceptions import cv_raise, ViewSetNotFoundError, ViewSetKeyFoundError
from crud_views.lib.viewset import path_regs
from .parentviewset import ParentViewSet
from .path_regs import PrimaryKeys
from .. import check
from ..check import CheckAttributeReg, Check
from ..settings import crud_views_settings
from ..view import (
    ContextButton,
    ParentContextButton,
    CrudView,
    ViewContext,
    CrudViewPermissionRequiredMixin as CrudViewPermissionRequiredMixin,
)
from ..view.buttons import FilterContextButton
from ..views.manage import ManageView

User = get_user_model()


def empty_dict() -> dict:
    return dict()


def context_buttons_default(*args, **kwargs) -> Any:
    return [
        ContextButton(
            key="home",
            key_target="list",
            # I don't think we need templates here because the title should be the one of the list page where it links to
        ),
        ParentContextButton(key="parent", key_target="list"),
        FilterContextButton(),
    ]


_REGISTRY = OrderedDict()
_REGISTRY_LOCK = threading.Lock()


class ViewSet(BaseModel):
    """
    Manages CRUD collections of views

    Default views are
    - list      : list collection of objects
    - detail    : detail view of an object
    - create    : create a new object
    - update    : update an existing object
    - delete    : delete an existing object

    Which ap to the following default permissions:
    - list      : view
    - detail    : view
    - create    : add
    - update    : change
    - delete    : delete
    """

    PK: ClassVar[Type[PrimaryKeys]] = PrimaryKeys

    model: Type[Model]
    name: str
    prefix: str | None = None
    app: str | None = None
    pk: str | None = None  # auto-detected from model if not set
    pk_name: str = "pk"
    context_buttons: List[ContextButton] = Field(default_factory=context_buttons_default)
    parent: ParentViewSet | None = None
    ordering: str | None = None
    icon_header: str | None = None

    _views: Dict[str, Type[CrudView]] = PrivateAttr(default_factory=empty_dict)  # noqa

    def __repr__(self):
        return f"ViewSet({self.name})"

    def __str__(self):
        return f"{self.name}"

    @model_validator(mode="after")
    def register(self) -> Self:
        if self.pk is None:
            from django.db import models

            pk_field_map = {
                models.UUIDField: self.PK.UUID,
                models.AutoField: self.PK.INT,
                models.BigAutoField: self.PK.INT,
                models.SmallAutoField: self.PK.INT,
                models.CharField: self.PK.STR,
                models.SlugField: self.PK.STR,
            }
            self.pk = pk_field_map.get(type(self.model._meta.pk), self.PK.INT)

        with _REGISTRY_LOCK:
            _REGISTRY[self.name] = self

        switch = crud_views_settings.manage_views_enabled
        if switch == "yes" or switch == "debug_only" and settings.DEBUG:

            class AutoManageView(ManageView):
                model = self.model
                cv_viewset = self

        return self

    @staticmethod
    def get_viewset(name: str) -> Self:
        if name not in _REGISTRY:
            raise ViewSetNotFoundError(name)
        return _REGISTRY[name]

    @staticmethod
    def checks_all() -> Iterable[Check]:
        """
        Iterator over all checks of all viewsets
        """
        with _REGISTRY_LOCK:
            for cv in _REGISTRY.values():
                yield from cv.checks()

    def has_view(self, name) -> bool:
        return name in self._views

    def get_all_views(self) -> Dict[str, Type[CrudView]]:
        return self._views

    def checks(self) -> Iterable[Check]:
        """
        Iterator over all checks
        """
        yield CheckAttributeReg(context=self, id="E002", attribute="name", **check.REGS["name"])
        yield CheckAttributeReg(context=self, id="E003", attribute="prefix", **check.REGS["path"])

        for view in self._views.values():
            yield from view.checks()

    @model_validator(mode="after")
    def model_validator_after(self) -> Self:
        if self.prefix is None:
            self.prefix = self.name
        return self

    def get_parent_url_args(self, first_only: bool = False) -> List[str] | str:
        """
        Get url args for all parents
        """
        args = []

        # iterate over parents
        if self.parent:
            parent = self.parent
            while parent is not None:
                arg = parent.get_pk_name()
                if first_only:
                    return arg
                args.append(parent.get_pk_name())
                parent = parent.viewset.parent

        return args

    @property
    def has_parent(self) -> bool:
        return self.parent is not None

    def get_parent_model(self):
        if self.parent:
            return self.parent.viewset.model
        return None

    def get_parent_attributes(self, first_only: bool = False) -> List[str] | str:
        """
        Get url args for all parents
        """
        attrs = []

        # iterate over parents
        if self.parent:
            parent = self.parent
            while parent is not None:
                attr = parent.get_attribute()
                if first_only:
                    return attr
                attrs.append(parent.get_attribute())
                parent = parent.viewset.parent

        return attrs

    def get_queryset(self, view: CrudView) -> QuerySet:
        """
        Queryset with respect to parent
        """

        q_kwargs = dict()
        if self.parent:
            parent = self.parent
            attribute = None
            while parent is not None:
                # parent args
                pk_name = parent.get_pk_name()
                if attribute is None:
                    attribute = parent.get_attribute()
                else:
                    attribute = f"{attribute}__{parent.get_attribute()}"

                # generate q-args
                q_kwargs.update({f"{attribute}__pk": view.kwargs[pk_name]})

                # proceed to next parent
                parent = parent.viewset.parent

            queryset = self.model.objects.filter(Q(**q_kwargs))

        else:
            # default queryset
            queryset = self.model.objects.all()

        # add ordering
        if self.ordering:
            queryset = queryset.order_by(self.ordering)

        return queryset

    def register_view_class(self, key: str, view_class: Type[CrudView]):
        cv_raise(key not in self._views, f"key {key} already registered at {self}")
        self._views[key] = view_class
        # add manage view to context
        if True:  # todo: based on settings
            if isinstance(view_class.cv_context_actions, list):
                if "manage" not in view_class.cv_context_actions and view_class.cv_key != "manage":
                    view_class.cv_context_actions.append("manage")

    def is_view_registered(self, key: str) -> bool:
        return key in self._views

    def get_view_class(self, key: str) -> Type[CrudView]:
        cv_raise(self.is_view_registered(key), f"key {key} not registered at {self}", ViewSetKeyFoundError)
        return self._views[key]

    def get_router_name(self, key: str) -> str:
        """
        The router name
        """
        app = f"{self.app}:" if self.app else ""
        return f"{app}{self.name}-{key}"

    def get_view_path(self, key: str) -> str:
        return self.get_view_class(key).cv_path

    def get_view_pk_path(self, key: str) -> str:
        if self.get_view_class(key).cv_object:
            return f"/{self.pk}"
        return ""

    def get_path_parent(self) -> str:
        """
        Return path_parent of URL pattern [path_parent/]path_prefix/[path_pk/]path_view/
        """
        paths = []

        parent = self.parent
        while parent:
            path_prefix = parent.viewset.get_path_prefix()
            path_pk = path_regs.get_path_pk(parent.get_pk_name(), parent.viewset.pk) + "/"
            path = f"{path_prefix}{path_pk}"
            paths.insert(0, path)
            parent = parent.viewset.parent

        path_parent = "".join(paths)
        return path_parent

    def get_path_prefix(self) -> str:
        """
        Return path_prefix of URL pattern [path_parent/]path_prefix/[path_pk/]path_view/
        """
        return f"{self.prefix}/"

    def get_path_pk(self, key: str) -> str:
        """
        Return path_pk of URL pattern [path_parent/]path_prefix/[path_pk/]path_view/
        """
        if self.get_view_class(key).cv_object:
            path_pk = path_regs.get_path_pk(self.pk_name, self.pk)
            return f"{path_pk}/"
        return ""

    def get_path_view(self, key: str) -> str:
        """
        Return path_view of URL pattern [path_parent/]path_prefix/[path_pk/]path_view/
        """
        path = self.get_view_class(key).cv_path
        if not path:
            return ""
        return f"{path}/"

    @cached_property
    def urlpatterns(self) -> List[URLResolver]:
        """
        Create urlpatterns for all views of ViewSet
        """

        # check if any views are registered
        cv_raise(len(self._views.keys()) > 0, f"no views registered at {self}")

        urlpatterns = []

        # common path parts
        path_parent = self.get_path_parent()
        path_prefix = self.get_path_prefix()

        # create urlpatterns for all views
        for key, view_class in self._views.items():
            # view specific path parts
            path_pk = self.get_path_pk(key)
            path_view = self.get_path_view(key)
            path_contribute = view_class.cv_path_contribute()

            # args for re_path
            route = rf"^{path_parent}{path_prefix}{path_pk}{path_view}{path_contribute}$"
            view = view_class.as_view()  # noqa
            name = self.get_router_name(key)

            # create and add URLResolver
            resolver = re_path(route, view, name=name)
            urlpatterns.append(resolver)

        return urlpatterns

    @cached_property
    def default_permissions(self) -> OrderedDict[str, str]:
        """
        Default permissions extracted from model
            - add
            - change
            - delete
            - view
            - ...
            - and custom permissions defined on model
        """
        model = self.model  # noqa
        content_type = ContentType.objects.get_for_model(model)
        permissions = OrderedDict()
        for permission in Permission.objects.filter(content_type=content_type):
            # todo: model name must in codename
            action = permission.codename.split(f"_{permission.content_type.model}")[0]
            permissions[action] = f"{permission.content_type.app_label}.{permission.codename}"
        return permissions

    @cached_property
    def permissions(self) -> OrderedDict[str, str]:
        # todo: move to view ?
        return self.default_permissions

    def get_meta(self, context: ViewContext) -> dict:
        """
        Meta data plus X for template context.
        """
        meta = self.model._meta
        data = {
            "viewset": self,
            "cv": self,
            "verbose_name": meta.verbose_name.capitalize(),
            "verbose_name_plural": meta.verbose_name_plural.capitalize(),
        }
        data.update(
            {
                "verbose_name_translate": _(data["verbose_name"]),
                "verbose_name_plural_translate": _(data["verbose_name_plural"]),
            }
        )
        data.update(context.to_dict())
        return data



from __future__ import annotations

from typing import Dict, Type, Iterable

from crud_views.lib.check import Check, CheckAttribute
from django.db.models import Model
from django.forms.models import ModelForm
from django.http import JsonResponse

from .formsets import FormSets


class FormSetMixinBase:
    cv_formsets_required: bool = True

    @classmethod
    def checks(cls) -> Iterable[Check]:
        """
        Iterator of checks for the view
        """
        yield from super().checks()
        yield CheckAttribute(context=cls, id="E200", attribute="cv_formsets_required")  # todo check for type

    def get_context_data(self, **kwargs):
        """
        In formset mixin formsets need to be created to be added to the context
        """
        data = super().get_context_data(**kwargs)  # noqa
        formsets = self.cv_init_formsets(data.get("form"))
        if formsets is not None:
            data["formsets"] = formsets
        return data

    def get(self, request, *args, **kwargs):
        """
        GET also handles AJAX requests for formset templates
        """
        template = request.GET.get("template", None)
        if template:
            data = self.get_template_html(request)
            return JsonResponse(data)
        return super().get(request, *args, **kwargs)  # noqa

    def get_template_html(self, request):

        key_path = request.GET.get("template").split("|")
        pk = request.GET.get("pk", "None")
        num = request.GET.get("num")
        formset_parent_prefix_key = request.GET.get("formset_parent_prefix_key")

        formsets = self.cv_get_formsets()
        data = formsets.get_template(key_path=key_path, pk=pk, num=num, parent_prefix=formset_parent_prefix_key)
        return data

    def cv_form_is_valid(self, context: dict) -> bool:
        """
        Check if the form is valid.
        Crud Views modules may extend this method with further checks.
        """

        # the main form
        form_valid = super().cv_form_is_valid(context)

        # get the formsets
        formsets = context.get("formsets", None)
        if formsets is None:
            if self.cv_formsets_required:
                raise ValueError("Formsets are required but not defined, cv_formsets_required=True")
            else:
                return form_valid

        # formsets valid?
        formsets_valid = list(formsets.is_valid())
        all_formsets_valid = all([v for fs, v in formsets_valid])

        # form and formsets valid?
        is_valid = all([form_valid, all_formsets_valid])
        return is_valid

    def cv_form_valid(self, context: dict):
        """
        Save form and formsets
        """
        # save main form
        super().cv_form_valid(context)

        # get the formsets
        formsets = context.get("formsets", None)
        if formsets is None:
            if self.cv_formsets_required:
                raise ValueError("Formsets are required but not defined, cv_formsets_required=True")

            # nothing to do here
            return

        # save formsets
        formsets.save(commit=True)

    def cv_get_formsets(self) -> FormSets:
        return self.cv_formsets.clone(cv_view=self)  # noqa

    def cv_patch_formsets(self, formsets: FormSets):
        pass

    def cv_init_formsets(self, form: ModelForm) -> FormSets | None:
        """
        Get the inline formsets instances.
        On POST the formsets are bound to the request.POST data.
        """

        # get inline formsets
        formsets = self.cv_get_formsets()

        # if no formsets are defined, return None
        if formsets is None:
            return None

        formsets.init(request=self.request, form=form, instance=self.object)

        formsets.init_js_data(self)

        # call hook to patch the formsets
        self.cv_patch_formsets(formsets)

        return formsets

    # def cv_form_valid(self, form: ModelForm, formsets: FormSets):
    #     """
    #     Use a custom form_valid method to handle formsets because the form_valid method
    #     would recreate the formsets again
    #     """
    #
    #     with transaction.atomic():
    #         # delete first
    #         # formsets.deleted()
    #
    #         self.object = form.save(commit=True)
    #         formsets.save(commit=True)
    #
    #     return HttpResponseRedirect(self.get_success_url())  # noqa


class FormSetMixin(FormSetMixinBase):
    cv_formsets: FormSets

    # todo: checks for cv_formsets: FormSets

    @classmethod
    def checks(cls) -> Iterable[Check]:
        """
        Iterator of checks for the view
        """
        yield from super().checks()
        yield CheckAttribute(context=cls, id="E200", attribute="cv_formsets")  # todo check for type

    def cv_get_formsets(self) -> FormSets:
        return self.cv_formsets.clone(cv_view=self)  # noqa


# todo: rename, because this class name collides with django polymorphic
class PolymorphicFormSetMixin(FormSetMixinBase):
    cv_polymorphic_formsets: Dict[Type[Model], FormSets]

    # todo: checks for cv_polymorphic_formsets

    @classmethod
    def checks(cls) -> Iterable[Check]:
        """
        Iterator of checks for the view
        """
        yield from super().checks()
        yield CheckAttribute(context=cls, id="E200", attribute="cv_polymorphic_formsets")  # todo check for type

    def cv_get_formsets(self) -> FormSets | None:
        model = self.polymorphic_model

        # it is okay that a model has no formsets defined
        if model not in self.cv_polymorphic_formsets:
            raise ValueError(f"No FormSets instance found for polymorphic model {model.__class__}")
        formsets = self.cv_polymorphic_formsets.get(model, None)
        if formsets is None:
            return None
        return formsets.clone(cv_view=self)  # noqa



import django_tables2 as tables
from crispy_forms.layout import Row

from app.models import Baz
from crud_views.lib.crispy import CrispyModelForm, Column4, CrispyModelViewMixin, CrispyDeleteForm
from crud_views.lib.table import Table, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
    ListViewPermissionRequired,
    CreateViewParentMixin,
)
from crud_views.lib.viewset import ViewSet, ParentViewSet

cv_baz = ViewSet(model=Baz, name="baz", parent=ParentViewSet(name="bar"), icon_header="fa-solid fa-dog")


class BazForm(CrispyModelForm):
    submit_label = "Create"

    class Meta:
        model = Baz
        fields = ["name"]

    def get_layout_fields(self):
        return Row(Column4("name"))


class BazTable(Table):
    id = LinkDetailColumn()
    name = tables.Column()

    def render_baz(self, record):
        return "baz"


class BazListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Baz
    table_class = BazTable
    cv_viewset = cv_baz
    cv_list_actions = ["detail", "update", "delete"]


class BazDetailView(DetailViewPermissionRequired):
    model = Baz
    cv_viewset = cv_baz
    cv_property_display = [
        {
            "title": "Properties",
            "icon": "dog",
            "description": "Baz attributes",
            "properties": [
                "id",
                "name",
            ],
        },
    ]


class BazUpdateView(CrispyModelViewMixin, UpdateViewPermissionRequired):
    model = Baz
    form_class = BazForm

    cv_viewset = cv_baz


class BazCreateView(CrispyModelViewMixin, CreateViewParentMixin, CreateViewPermissionRequired):
    model = Baz
    form_class = BazForm
    cv_viewset = cv_baz


class BazDeleteView(CrispyModelViewMixin, DeleteViewPermissionRequired):
    model = Baz
    form_class = CrispyDeleteForm
    cv_viewset = cv_baz



from functools import cached_property
from typing import List, Dict, Generator, Type

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model
from django.forms import ModelForm
from polymorphic.models import PolymorphicModel


def get_polymorphic_child_models(model: PolymorphicModel) -> List[Type[PolymorphicModel]]:
    """
    Get all child models of a polymorphic model.
    """
    assert issubclass(model, PolymorphicModel), "not a polymorphic model"

    def subclasses(m: Type[PolymorphicModel]) -> Generator[Type[PolymorphicModel], None, None]:
        for sm in m.__subclasses__():
            yield sm
            yield from subclasses(sm)

    return list(subclasses(model))


def get_polymorphic_child_models_content_types(model: PolymorphicModel) -> Dict[Type[Model], ContentType]:
    child_models = get_polymorphic_child_models(model)
    content_types = ContentType.objects.get_for_models(*child_models, for_concrete_models=True)
    return content_types


class PolymorphicCrudViewMixin:
    """
    Polymorphic ViewSet mixin
    """

    polymorphic_forms: Dict[Model, ModelForm] = None

    @property
    def polymorphic_ctype_id(self) -> int:
        """
        Get the polymorphic content type id from the url kwargs
        """
        polymorphic_ctype_id = self.kwargs.get("polymorphic_ctype_id")
        if not polymorphic_ctype_id:
            instance = self.get_object()  # from SingleObjectMixin
            polymorphic_ctype_id = instance.polymorphic_ctype_id
        return int(polymorphic_ctype_id)

    @cached_property
    def polymorphic_model(self) -> Model:
        """
        Get the polymorphic model from polymorphic_ctype_id
        """
        polymorphic_ctype_id = self.polymorphic_ctype_id
        content_type = ContentType.objects.get(id=polymorphic_ctype_id)
        return content_type.model_class()

    @property
    def model(self) -> Model:
        """
        Override the model property from the view mixin
        """
        return self.polymorphic_model

    def get_form_class(self):
        """
        Get the form class depending on the polymorphic model
        """
        model = self.polymorphic_model
        form = self.polymorphic_forms.get(model, None)
        if not form:
            raise ValueError(f"No form found for polymorphic model {self.object.__class__}")
        return form



from .table import (
    Column1,
    Column2,
    Column3,
    Column4,
    Column5,
    Column6,
    Column7,
    Column8,
    Column9,
    Column10,
    Column11,
    Column12,
)
from .form import CrispyModelForm, CrispyViewMixin, CrispyModelViewMixin, CrispyForm, CrispyDeleteForm

__all__ = [
    "Column1",
    "Column2",
    "Column3",
    "Column4",
    "Column5",
    "Column6",
    "Column7",
    "Column8",
    "Column9",
    "Column10",
    "Column11",
    "Column12",
    "CrispyModelForm",
    "CrispyForm",
    "CrispyDeleteForm",
    "CrispyModelViewMixin",
    "CrispyViewMixin",
]



import json
from typing import Iterable
from urllib.parse import parse_qs

from django.contrib import messages
from django.core.exceptions import BadRequest
from django.http import HttpResponseRedirect
from django.http import JsonResponse
from django_filters.views import FilterView
from django_tables2 import SingleTableMixin

from crud_views.lib.check import Check, CheckTemplateOrCode
from crud_views.lib.session import SessionData
from crud_views.lib.settings import crud_views_settings


class CrudViewProcessFormMixin:
    """
    Mixin for create and update views.
    Note ProcessFormView.post is overridden.
    Why? Because we need a more detailed handling of the post method.
    """

    def post(self, request, *args, **kwargs):
        """
        Override ProcessFormView
        """

        try:
            self.object = self.get_object()  # noqa
        except AttributeError:
            self.object = None  # noqa

        context = self.get_context_data(**kwargs)
        self.cv_post_hook(context)
        if self.cv_form_is_valid(context):
            self.cv_form_valid(context)
            self.cv_form_valid_hook(context)
            return self.cv_form_valid_redirect(context)
        else:
            self.cv_form_invalid_hook(context)
            return self.cv_form_invalid(context)

    def cv_post_hook(self, context: dict):
        """
        Hook on post.
        Crud Views modules may extend this method.
        """
        pass

    def cv_form_is_valid(self, context: dict) -> bool:
        """
        Check if the form is valid.
        Crud Views modules may extend this method with further checks.
        """
        return context["form"].is_valid()

    def cv_form_valid(self, context: dict):
        """
        Handle valid form
        """
        pass

    def cv_form_valid_hook(self, context: dict):
        """
        Handle valid form hook
        """
        pass

    def cv_form_invalid(self, context: dict):
        """
        Handle invalid form
        """
        return self.render_to_response(context)

    def cv_form_invalid_hook(self, context: dict):
        """
        Handle invalid form hook
        """
        pass

    def cv_form_valid_redirect(self, context: dict) -> HttpResponseRedirect:
        """
        Redirect to the success url
        """
        return HttpResponseRedirect(self.get_success_url())

    def cv_form_invalid_redirect(self, context: dict) -> HttpResponseRedirect:
        """
        Handle invalid form
        """
        return self.render_to_response(context)


class MessageMixin:
    """
    Add messages for a view.
    Note: the view must configure the message template and message template or code:
            - cv_message_template
            - cv_message_template_code
    """

    @classmethod
    def checks(cls) -> Iterable[Check]:
        """
        Iterator of checks for the view
        """
        yield from super().checks()  # noqa
        yield CheckTemplateOrCode(context=cls, attribute="cv_message_template")

    def cv_get_message(self, attribute: str = "cv_message") -> str | None:
        return self.render_snippet(
            self.cv_get_meta(),
            self.cv_message_template,
            self.cv_message_template_code,
        )

    def cv_form_valid_hook(self, context: dict):
        super().cv_form_valid_hook(context)  # noqa
        message = self.cv_get_message()
        if message:
            messages.success(self.request, message)

    def action(self, context: dict) -> bool:
        result = super().action(context)
        if result:
            messages.success(self.request, self.cv_get_message())
        elif hasattr(self, "cv_error_message"):
            # error message is optional
            messages.error(self.request, self.cv_get_message("cv_error_message"))

        return result


class ListViewTableMixin(SingleTableMixin):
    """
    Mixin for ListView to render tables with django-tables2
    """

    template_name = "crud_views/view_list_table.html"

    table: SingleTableMixin = None
    table_class: str = None
    paginate_by: int = 10

    def get_table_kwargs(self):
        return {"view": self}


class ListViewTableFilterMixin(FilterView):
    """
    Mixin for ListView to add a filter with django-filter.
    Additionally, there is a filter persistence for the fields and the expanded state of the filter.
    Persistence ist store in the user session.

    Note: not all themes may support the expanded state of the filter.
    """

    filterset_class = None
    formhelper_class = None

    cv_filter_persistence: bool = crud_views_settings.filter_persistence
    cv_session_key_querystring: str = "filter_query_string"

    # todo: check cv_filter_persistence cv_session_key_querystring

    def get_filterset(self, filterset_class):  # noqa
        kwargs = self.get_filterset_kwargs(filterset_class)
        filterset = filterset_class(**kwargs)
        if self.formhelper_class:
            filterset.form.helper = self.formhelper_class(request=self.request)
        return filterset

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        filter_expanded = SessionData.from_view(self).get("filter_expanded", False)
        #
        # with SessionData.from_view(self) as sd:
        #     filter_expanded = sd["filter_expanded"]

        context["cv_filter_expanded"] = filter_expanded
        return context

    def post(self, request, *args, **kwargs):
        """
        Store filter expanded state
        """

        if not self.cv_filter_persistence:
            return JsonResponse({"status": "ok", "filter_expanded": None})

        data = json.loads(request.body.decode("utf-8"))
        filter_expanded = data.get("filter_expanded", None)
        if filter_expanded is None:
            raise BadRequest("filter_expanded not set")

        with SessionData.from_view(self) as sd:
            sd["filter_expanded"] = filter_expanded

        return JsonResponse({"status": "ok", "filter_expanded": filter_expanded})

    def get(self, request, *args, **kwargs):
        """
        Add filter persistence via session here.
        This was inspired by Lorenzo Prodon's django-persistent-filters:
            https://github.com/LorenzoProd/django-persistent-filters
        """

        if not self.cv_filter_persistence:
            return super().get(request, *args, **kwargs)

        with SessionData.from_view(self) as sd:
            # get stored query string and request query string
            stored_query_string = sd.get(self.cv_session_key_querystring, "")
            query_string = self.request.META["QUERY_STRING"]

            # reset filter ?
            qs = parse_qs(query_string)  # todo: is this correct? looks strange with that lists
            reset_filter = qs.get("reset_filter", ["false"])[0] == "true"
            if reset_filter:
                try:
                    del sd[self.cv_session_key_querystring]
                except KeyError:
                    pass
                url = self.request.path
                sort = qs.get("sort", [None])[0]
                if sort:
                    url += f"?sort={sort}"
                return HttpResponseRedirect(url)

            # there is a query string, update data
            if len(query_string) > 0:
                sd[self.cv_session_key_querystring] = query_string

            # no query string, but data in session, restore query string
            if len(query_string) == 0 and len(stored_query_string):
                redirect_to = self.request.path + "?" + stored_query_string
                return HttpResponseRedirect(redirect_to)

        return super().get(request, *args, **kwargs)



import django_tables2 as tables
from crispy_forms.layout import Row
from django.utils.translation import gettext as _

from app.models import Group
from crud_views.lib.crispy import Column4, CrispyModelForm, CrispyModelViewMixin, CrispyDeleteForm
from crud_views.lib.table import Table, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    MessageMixin,
    ListViewTableMixin,
    ListViewPermissionRequired,
    DeleteViewPermissionRequired,
    RedirectChildView,
)
from crud_views.lib.viewset import ViewSet

cv_group = ViewSet(model=Group, name="group", icon_header="fa-solid fa-user-group")


class GroupCreateForm(CrispyModelForm):
    submit_label = _("Create")

    class Meta:
        model = Group
        fields = ["name"]

    def get_layout_fields(self):
        return Row(Column4("name"))


class GroupUpdateForm(GroupCreateForm):
    """
    Update form has the same fields as the create form
    """

    submit_label = _("Update")


class GroupTable(Table):
    id = LinkDetailColumn(attrs=Table.ca.ID)
    name = tables.Column()


class GroupListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Group

    cv_viewset = cv_group
    cv_list_actions = ["detail", "update", "delete", "redirect_child"]

    table_class = GroupTable


class GroupCreateView(CrispyModelViewMixin, MessageMixin, CreateViewPermissionRequired):
    model = Group
    form_class = GroupCreateForm
    cv_viewset = cv_group
    cv_message = "Created author »{object}«"


class GroupUpdateView(CrispyModelViewMixin, MessageMixin, UpdateViewPermissionRequired):
    model = Group
    form_class = GroupUpdateForm
    cv_viewset = cv_group
    cv_message = "Updated author »{object}«"


class GroupDeleteView(CrispyModelViewMixin, MessageMixin, DeleteViewPermissionRequired):
    model = Group
    form_class = CrispyDeleteForm
    cv_viewset = cv_group
    cv_message = "Deleted author »{object}«"


class GroupDetailView(DetailViewPermissionRequired):
    model = Group
    cv_viewset = cv_group

    cv_property_display = [
        {
            "title": _("Attributes"),
            "icon": "users",
            "description": _("Group identification"),
            "properties": [
                "id",
                "name",
            ],
        },
    ]


class RedirectMembersView(RedirectChildView):
    cv_action_label = _("Manage Members")
    cv_redirect = "members"
    cv_redirect_key = "list"
    cv_icon_action = "fa-regular fa-user"

    cv_viewset = cv_group



from typing import List

from lxml import html
from lxml.html import Element


class Header:
    def __init__(self, element: Element):
        self.element = element

    @property
    def text(self) -> bool:
        return self.element.text_content().strip()

    @property
    def orderable(self) -> bool:
        return "orderable" in self.element.classes

    @property
    def is_action(self) -> bool:
        return self.text == "Actions"

    @property
    def link(self):
        if self.is_action:
            return None

    @property
    def actions(self):
        if self.is_action:
            for index, element in self.element.xpath("a"):
                yield Action(index, element)


class Action:
    def __init__(self, index: int, element: Element):
        self.index = index
        self.element = element

    def __str__(self):
        try:
            key = self.key
        except Exception:
            key = "KEY-ERROR"
        return f"action({key}:{self.index}:{self.text})"

    @property
    def is_disabled(self) -> bool:
        if self.element.tag == "span":
            return "disabled" in self.element.xpath("a")[0].classes
        return "disabled" in self.element.classes

    @property
    def text(self) -> str:
        return self.element.text_content().strip()

    @property
    def title(self) -> str:
        """
        Get the title either from span or from a tag
        """
        return self.element.attrib["title"].strip()

    @property
    def key(self) -> str | None:
        if self.element.tag == "span":
            return self.element.xpath("a")[0].attrib.get("cv-key", None)
        return self.element.attrib.get("cv-key", None)

    @property
    def href(self) -> str | None:
        # disabled buttons have no href
        if self.element.tag == "span":
            return self.element.xpath("a")[0].attrib.get("href")
        # context actions do not have a span when disabled
        return self.element.attrib.get("href")


class Column:
    def __init__(self, index: int, row: "Row", element: Element, action_index: int):
        self.index = index
        self.row = row
        self.element = element
        self.action_index = action_index

    def __str__(self):
        return f"col({self.row.index},{self.index}:{self.text})"

    @property
    def text(self) -> bool:
        return self.element.text_content().strip()

    @property
    def actions(self) -> List[Action]:
        assert self.index == self.action_index
        data: List[Action] = []
        for index, element in enumerate(self.element.xpath("div/a|span")):
            action = Action(index, element)
            data.append(action)
        return data


class Row:
    def __init__(self, index: int, element: Element, action_index: int):
        self.index = index
        self.element = element
        self.action_index = action_index

    def __str__(self):
        return f"row({self.index})"

    @property
    def columns(self) -> List[Column]:
        result = []
        for index, element in enumerate(self.element.xpath("td")):
            column = Column(index, self, element, action_index=self.action_index)
            result.append(column)
        return result

    @property
    def actions(self) -> List[Action]:
        column = self.columns[self.action_index]
        return column.actions

    def get_action(self, key: str) -> Action:
        for action in self.actions:
            if action.key == key:
                return action
        raise KeyError(key)


class Table:
    def __init__(self, response):
        self.content = response.content.decode("utf-8").strip()
        self.html = html.fromstring(self.content)

    @property
    def context(self) -> Element:
        return self.html.xpath("//div[@cv-context-container='true']")[0]

    @property
    def context_actions(self) -> List[Action]:
        return [Action(index=index, element=element) for index, element in enumerate(self.context.xpath("a"))]

    def get_context_action(self, key: str) -> Action:
        for action in self.context_actions:
            if action.key == key:
                return action
        raise KeyError(key)

    @property
    def table(self):
        return self.html.xpath("//table")[0]

    @property
    def tbody(self):
        return self.table.xpath("//tbody")[0]

    @property
    def rows(self) -> List[Row]:
        action_index = self.action_index
        return [Row(index, row, action_index=action_index) for index, row in enumerate(self.tbody.xpath("tr"))]

    @property
    def thead(self):
        return self.table.xpath("//thead")[0]

    @property
    def headers(self) -> List[Header]:
        return [Header(header) for header in self.thead.xpath("//th")]

    @property
    def action_index(self) -> int | None:
        for index, header in enumerate(self.headers):
            if header.is_action:
                return index



from .table import Table, TableWithViewContext
from .columns import (
    ActionColumn,
    LinkChildColumn,
    LinkDetailColumnMixin,
    LinkDetailColumn,
    UUIDColumn,
    UUIDLinkDetailColumn,
)

__all__ = [
    "Table",
    "TableWithViewContext",
    "ActionColumn",
    "LinkChildColumn",
    "LinkDetailColumnMixin",
    "LinkDetailColumn",
    "UUIDColumn",
    "UUIDLinkDetailColumn",
]



# Generated by Django 5.1.7 on 2025-04-16 18:05

import django.db.models.deletion
import uuid
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Author",
            fields=[
                (
                    "order",
                    models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order"),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "first_name",
                    models.CharField(max_length=100, verbose_name="Vorname"),
                ),
                ("last_name", models.CharField(max_length=100)),
                ("pseudonym", models.CharField(blank=True, max_length=100, null=True)),
                ("created_dt", models.DateTimeField(auto_now_add=True)),
                ("modified_dt", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "Autor",
                "verbose_name_plural": "Autoren",
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Bar",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="Foo",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
            ],
        ),
        migrations.CreateModel(
            name="Parent",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=100)),
            ],
            options={
                "ordering": ["name"],
            },
        ),
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
            ],
        ),
        migrations.CreateModel(
            name="Poly",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("shared", models.CharField(max_length=100, null=True)),
                (
                    "polymorphic_ctype",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="polymorphic_%(app_label)s.%(class)s_set+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "ordering": ["shared"],
            },
        ),
        migrations.CreateModel(
            name="PolyAnswer",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "poly",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.poly"),
                ),
                (
                    "polymorphic_ctype",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="polymorphic_%(app_label)s.%(class)s_set+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
        ),
        migrations.CreateModel(
            name="PolyTwoChoice",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order"),
                ),
                ("choice", models.CharField(max_length=100)),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Baz",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                (
                    "bar",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.bar"),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Book",
            fields=[
                (
                    "order",
                    models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order"),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("title", models.CharField(max_length=100)),
                ("price", models.DecimalField(decimal_places=2, max_digits=10)),
                ("created_dt", models.DateTimeField(auto_now_add=True)),
                ("modified_dt", models.DateTimeField(auto_now=True)),
                (
                    "author",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.author"),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Detail",
            fields=[
                (
                    "order",
                    models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order"),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("integer", models.IntegerField(verbose_name="An integer field")),
                ("number", models.FloatField(verbose_name="A float field")),
                ("char", models.CharField(max_length=100, verbose_name="Text field")),
                ("text", models.TextField(verbose_name="Multiline Text")),
                (
                    "boolean",
                    models.BooleanField(default=None, null=True, verbose_name="A boolean value"),
                ),
                (
                    "boolean_two",
                    models.BooleanField(default=None, null=True, verbose_name="Another boolean value"),
                ),
                ("date", models.DateField(verbose_name="A date field")),
                (
                    "date_time",
                    models.DateTimeField(verbose_name="A date field with time"),
                ),
                ("created_dt", models.DateTimeField(auto_now_add=True)),
                ("modified_dt", models.DateTimeField(auto_now=True)),
                (
                    "author",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="app.author",
                        verbose_name="A foreign key field",
                    ),
                ),
                (
                    "foo",
                    models.ManyToManyField(to="app.foo", verbose_name="Foo selected"),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="bar",
            name="foo",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.foo"),
        ),
        migrations.CreateModel(
            name="Membership",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date_joined", models.DateField()),
                ("invite_reason", models.CharField(max_length=64)),
                (
                    "group",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.group"),
                ),
                (
                    "person",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.person"),
                ),
            ],
        ),
        migrations.AddField(
            model_name="group",
            name="members",
            field=models.ManyToManyField(related_name="group", through="app.Membership", to="app.person"),
        ),
        migrations.CreateModel(
            name="PolyOne",
            fields=[
                (
                    "poly_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="app.poly",
                    ),
                ),
                ("one", models.CharField(max_length=100)),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            bases=("app.poly",),
        ),
        migrations.CreateModel(
            name="PolyThree",
            fields=[
                (
                    "poly_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="app.poly",
                    ),
                ),
                ("three", models.CharField(max_length=100)),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            bases=("app.poly",),
        ),
        migrations.CreateModel(
            name="PolyTwo",
            fields=[
                (
                    "poly_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="app.poly",
                    ),
                ),
                ("two", models.CharField(max_length=100)),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            bases=("app.poly",),
        ),
        migrations.CreateModel(
            name="PolyAnswerNumber",
            fields=[
                (
                    "polyanswer_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="app.polyanswer",
                    ),
                ),
                ("answer", models.IntegerField(null=True)),
                ("number", models.IntegerField(null=True)),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            bases=("app.polyanswer",),
        ),
        migrations.CreateModel(
            name="PolyAnswerText",
            fields=[
                (
                    "polyanswer_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="app.polyanswer",
                    ),
                ),
                ("answer", models.CharField(max_length=100, null=True)),
                ("text", models.CharField(max_length=100, null=True)),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            bases=("app.polyanswer",),
        ),
        migrations.CreateModel(
            name="PolyParent",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date_joined", models.DateField()),
                (
                    "parent",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.parent"),
                ),
                (
                    "poly",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.poly"),
                ),
            ],
        ),
        migrations.AddField(
            model_name="parent",
            name="polys",
            field=models.ManyToManyField(related_name="parents", through="app.PolyParent", to="app.poly"),
        ),
        migrations.AddConstraint(
            model_name="membership",
            constraint=models.UniqueConstraint(fields=("person", "group"), name="unique_person_group"),
        ),
        migrations.AddField(
            model_name="polytwochoice",
            name="poly",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.polytwo"),
        ),
        migrations.AddConstraint(
            model_name="polyparent",
            constraint=models.UniqueConstraint(fields=("poly", "parent"), name="unique_poly_parent"),
        ),
    ]



import pytest
from django.contrib.auth import get_user_model
from django.test import RequestFactory

from tests.test1.app.views import AuthorListView
from crud_views.lib.viewset import ViewSet

User = get_user_model()


def instantiate_view(cls, request, **initkwargs):
    """Instantiate a view class via RequestFactory without dispatching."""
    view = cls(**initkwargs)
    view.setup(request)
    return view


@pytest.mark.django_db
def test_factory_success_url(user_author_view: User, cv_author: ViewSet, author_douglas_adams):
    factory = RequestFactory()
    request = factory.get("/author/")
    request.user = user_author_view

    view = instantiate_view(AuthorListView, request)
    success_url = view.get_success_url()

    assert success_url == "/author/"



import re
from typing import Any, Iterable, Type

from django.core.checks import Error, CheckMessage
from django.template import TemplateDoesNotExist
from django.template.loader import get_template
from pydantic import BaseModel

REGS = {
    "name": {"reg": re.compile(r"^[a-z_]+$"), "msg": "must be lowercase alpha with underscores"},
    "path": {
        "reg": re.compile(r"^$|^[a-z0-9_-]+(?:/[a-z0-9_-]+)*$"),
        "msg": "must be lowercase alpha with underscores and dashes and must not start or end with a slash",
    },
}


class Check(BaseModel):
    """
    Base class for checks
    """

    context: Type | object
    id: str
    msg: str | None = None

    def get_id(self) -> str:
        """
        error id
        """
        return f"viewset.{self.id}"

    def get_message_context(self) -> dict:
        return dict(
            context=self.context,
            id=self.id,
            eid=self.get_id(),
        )

    def get_message(self, msg: str = None) -> str:
        kwargs = self.get_message_context()
        if msg:
            msg.format(**kwargs)
        return self.msg.format(**kwargs)


# todo: check for type
class CheckAttribute(Check):
    """
    Check for attribute
    """

    id: str = "E100"
    attribute: str | None = None
    nullable: bool = False
    msg: str = "Attribute »{attribute}» does not exist or is not set at »{context}»"

    def get_message_context(self) -> dict:
        context = super().get_message_context()
        context.update(
            attribute=self.attribute,
            value=self.value,
        )
        return context

    @property
    def exists(self) -> bool:
        return hasattr(self.context, self.attribute)

    @property
    def value(self) -> Any:
        return getattr(self.context, self.attribute)

    def messages(self) -> Iterable[CheckMessage]:
        # yield from super().messages()
        if not self.exists or (self.value is None and not self.nullable):
            yield Error(id=self.get_id(), msg=self.get_message())


class CheckAttributeReg(CheckAttribute):
    """
    Check attribute against regex
    """

    reg: re.Pattern
    msg: str = "Attribute »{attribute}» value »{value}» does not match regex »{reg}» at »{context}»"

    def get_message_context(self) -> dict:
        context = super().get_message_context()
        context.update(
            reg=self.reg,
        )
        return context

    def messages(self) -> Iterable[CheckMessage]:
        yield from super().messages()
        if self.exists and not self.reg.match(self.value):
            yield Error(id=f"viewset.{self.id}", msg=self.get_message())


class CheckEitherAttribute(Check):
    """
    Check for either attribute
    """

    attribute1: str | None = None
    attribute2: str | None = None
    allow_none: bool = False

    msg: str = "Neither »{attribute1}» not »{attribute2}» are set or are missing »{context}»"

    def get_message_context(self) -> dict:
        context = super().get_message_context()
        context.update(
            attribute1=self.attribute1,
            attribute2=self.attribute2,
            value1=self.value1,
            value2=self.value2,
        )
        return context

    @property
    def value1(self) -> Any:
        return getattr(self.context, self.attribute1, None)

    @property
    def value2(self) -> Any:
        return getattr(self.context, self.attribute2, None)

    def messages(self) -> Iterable[CheckMessage]:
        # yield from super().messages()

        if not self.value1 and not self.value2 and not self.allow_none:
            yield Error(
                id=self.get_id(),
                msg=self.get_message(
                    "Neither attribute »{attribute1}» nor attribute »{attribute2}» are set at »{context}»"
                ),
            )

        elif self.value1 and self.value2:
            yield Error(
                id=self.get_id(),
                msg=self.get_message("Both attributes »{attribute1}» and »{attribute2}» are set at »{context}»"),
            )


class CheckTemplateOrCode(Check):
    """
    Check for template or template_code
    """

    id: str = "E110"
    attribute: str | None = None
    msg_none: str = "Neither »{attr_template}» nor »{attr_code}» defined at »{context}»"
    msg_template_not_found: str = "Template »{template}» not found at »{context}»"

    def get_message_context(self) -> dict:
        context = super().get_message_context()
        context.update(
            attribute=self.attribute,
        )
        return context

    def messages(self) -> Iterable[CheckMessage]:
        attr_template = f"{self.attribute}"
        attr_code = f"{self.attribute}_code"
        template = getattr(self.context, attr_template, None)
        code = getattr(self.context, attr_code, None)
        if not (template or code):
            msg = self.msg_none.format(attr_template=attr_template, attr_code=attr_code, context=self.context)
            yield Error(id=self.get_id(), msg=msg)
        if template:
            try:
                get_template(template)
            except TemplateDoesNotExist:
                msg = self.msg_template_not_found.format(template=template, context=self.context)
                yield Error(id=self.get_id(), msg=msg)


class ContextActionCheck(Check):
    """
    Checks for context action
    """

    # todo: check when all apps are loaded
    msg: str = "Attribute »{attribute}» does not exist or is not set at »{context}»"

    def messages(self) -> Iterable[CheckMessage]:
        viewset = self.context.cv  # noqa
        actions = self.context.cv_context_actions or list()  # noqa
        for action in actions:  # noqa
            is_view = viewset.has_view(action)
            # is_special = action in viewset.special_keys
            # if not (is_view or is_special):
            #    yield Error(id=f"viewset.{self.id}", msg=f"{self.msg} at {self.context}: {action}")
            if not is_view:
                yield Error(id=f"viewset.{self.id}", msg=f"{self.msg} at {self.context}: {action}")


class CheckExpression(Check):
    expression: bool
    msg: str = "foo"

    def messages(self) -> Iterable[CheckMessage]:
        if not self.expression:
            yield Error(id=f"viewset.{self.id}", msg=f"{self.msg} at {self.context}")



import pytest
from django.test.client import Client

from tests.lib.helper.boostrap5 import Table


@pytest.mark.django_db
def test_filter_no_params_shows_all(
    client_user_publisher_view: Client, cv_publisher, publisher_penguin, publisher_harpercollins
):
    """Without filter params, all publishers are listed."""
    response = client_user_publisher_view.get("/publisher/")
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 2


@pytest.mark.django_db
def test_filter_by_name_match(
    client_user_publisher_view: Client, cv_publisher, publisher_penguin, publisher_harpercollins
):
    """Filter with matching name returns only the matched publisher."""
    response = client_user_publisher_view.get("/publisher/?name=penguin")
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 1
    assert "Penguin" in table.rows[0].columns[1].text


@pytest.mark.django_db
def test_filter_by_name_no_match(
    client_user_publisher_view: Client, cv_publisher, publisher_penguin, publisher_harpercollins
):
    """Filter with non-matching name returns empty table."""
    response = client_user_publisher_view.get("/publisher/?name=nonexistent")
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 0


@pytest.mark.django_db
def test_filter_is_case_insensitive(client_user_publisher_view: Client, cv_publisher, publisher_penguin):
    """icontains lookup should be case-insensitive."""
    response = client_user_publisher_view.get("/publisher/?name=PENGUIN")
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 1
    assert "Penguin" in table.rows[0].columns[1].text


@pytest.mark.django_db
def test_filter_partial_match(
    client_user_publisher_view: Client, cv_publisher, publisher_penguin, publisher_harpercollins
):
    """icontains lookup should match partial strings."""
    response = client_user_publisher_view.get("/publisher/?name=harper")
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 1
    assert "HarperCollins" in table.rows[0].columns[1].text


@pytest.mark.django_db
def test_filter_form_rendered(client_user_publisher_view: Client, cv_publisher, publisher_penguin):
    """The filter form should be present in the response."""
    response = client_user_publisher_view.get("/publisher/")
    content = response.content.decode("utf-8")
    assert 'id="filter-button"' in content



import uuid

from django.db import models
from ordered_model.models import OrderedModel
from polymorphic.models import PolymorphicModel


# TODO: from ordered_model.models import OrderedModel


class Parent(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100)
    polys = models.ManyToManyField("Poly", through="PolyParent", related_name="parents")

    class Meta:
        ordering = ["name"]

    def __str__(self):
        return self.name


class Poly(PolymorphicModel, models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    shared = models.CharField(max_length=100, null=True)

    class Meta:
        ordering = ["shared"]


class PolyOne(Poly):
    one = models.CharField(max_length=100)


class PolyTwo(Poly):
    two = models.CharField(max_length=100)


class PolyThree(Poly):
    three = models.CharField(max_length=100)


class PolyTwoChoice(OrderedModel):
    poly = models.ForeignKey(PolyTwo, on_delete=models.CASCADE)
    choice = models.CharField(max_length=100)


class PolyAnswer(PolymorphicModel, models.Model):
    poly = models.ForeignKey(Poly, on_delete=models.CASCADE)


class PolyAnswerText(PolyAnswer):
    answer = models.CharField(max_length=100, null=True)
    text = models.CharField(max_length=100, null=True)


class PolyAnswerNumber(PolyAnswer):
    answer = models.IntegerField(null=True)
    number = models.IntegerField(null=True)


class PolyParent(models.Model):
    poly = models.ForeignKey(Poly, on_delete=models.CASCADE)
    parent = models.ForeignKey(Parent, on_delete=models.CASCADE)
    date_joined = models.DateField()

    class Meta:
        constraints = [models.UniqueConstraint(fields=["poly", "parent"], name="unique_poly_parent")]



from .formsets import FormSets as FormSets, FormSet as FormSet
from .mixins import FormSetMixin as FormSetMixin
from .layout import Formsets as Formsets, FormControl as FormControl
from .inline_formset import InlineFormSet as InlineFormSet



# Generated by Django 5.1.4 on 2025-01-20 07:24

import django.db.models.deletion
import uuid
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("app", "0002_poly_polyone_polytwo"),
    ]

    operations = [
        migrations.AlterField(
            model_name="author",
            name="first_name",
            field=models.CharField(max_length=100, verbose_name="Vorname"),
        ),
        migrations.CreateModel(
            name="Detail",
            fields=[
                ("order", models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order")),
                ("id", models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ("integer", models.IntegerField(verbose_name="An integer field")),
                ("number", models.FloatField(verbose_name="A float field")),
                ("char", models.CharField(max_length=100, verbose_name="Text field")),
                ("text", models.TextField(verbose_name="Multiline Text")),
                ("boolean", models.BooleanField(default=None, null=True, verbose_name="A boolean value")),
                ("boolean_two", models.BooleanField(default=None, null=True, verbose_name="Another boolean value")),
                ("date", models.DateField(verbose_name="A date field")),
                ("date_time", models.DateTimeField(verbose_name="A date field with time")),
                ("created_dt", models.DateTimeField(auto_now_add=True)),
                ("modified_dt", models.DateTimeField(auto_now=True)),
                (
                    "author",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="app.author",
                        verbose_name="A foreign key field",
                    ),
                ),
                ("foo", models.ManyToManyField(to="app.foo", verbose_name="Foo selected")),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
    ]



from .action_ordered import (
    OrderedUpView,
    OrderedDownView,
    OrderedUpViewPermissionRequired,
    OrderedUpDownPermissionRequired,
)
from .child import RedirectChildView
from .delete import DeleteView, DeleteViewPermissionRequired
from .list import ListView, ListViewPermissionRequired
from .detail import DetailView, DetailViewPermissionRequired
from .create import CreateView, CreateViewPermissionRequired, CreateViewParentMixin
from .mixins import MessageMixin, ListViewTableMixin, ListViewTableFilterMixin
from .update import UpdateView, UpdateViewPermissionRequired
from .action import ActionView, ActionViewPermissionRequired

__all__ = [
    # basic crud views
    "ListView",
    "ListViewPermissionRequired",
    "DetailView",
    "DetailViewPermissionRequired",
    "CreateView",
    "CreateViewParentMixin",
    "CreateViewPermissionRequired",
    "UpdateView",
    "UpdateViewPermissionRequired",
    "DeleteView",
    "DeleteViewPermissionRequired",
    "ActionView",
    "ActionViewPermissionRequired",
    # ordered
    "OrderedUpView",
    "OrderedUpViewPermissionRequired",
    "OrderedDownView",
    "OrderedUpDownPermissionRequired",
    # child
    "RedirectChildView",
    # mixins
    "MessageMixin",
    "ListViewTableMixin",
    "ListViewTableFilterMixin",
]



from crispy_forms.bootstrap import FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit, Layout
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.views import LoginView, LogoutView
from django.urls import path

from app.views import IndexView
from app.views.author import cv_author
from app.views.book import cv_book
from app.views.foo import cv_foo
from app.views.bar import cv_bar
from app.views.baz import cv_baz
from app.views.poly import cv_poly
from app.views.detail import cv_detail
from app.views.group import cv_group
from app.views.group_members import cv_person
from app.views.formset import cv_poly_formset
from app.views.formset.parent import cv_poly_parent_formset
from app.views.formset.question import cv_question
from app.views.campaign import cv_campaign


class CrispyAuthenticationForm(AuthenticationForm):
    @property
    def helper(self):
        helper = FormHelper()
        helper.layout = Layout("username", "password", FormActions(Submit("login", "Log In")))
        return helper


urlpatterns = [
    path("", IndexView.as_view(), name="index"),
    path("login/", LoginView.as_view(form_class=CrispyAuthenticationForm), name="login"),
    path("logout/", LogoutView.as_view(), name="logout"),
]

urlpatterns += (
    cv_author.urlpatterns
    + cv_book.urlpatterns
    + cv_foo.urlpatterns
    + cv_bar.urlpatterns
    + cv_baz.urlpatterns
    + cv_campaign.urlpatterns
    + cv_poly.urlpatterns
    + cv_detail.urlpatterns
    + cv_group.urlpatterns
    + cv_person.urlpatterns
    + cv_poly_formset.urlpatterns
    + cv_poly_parent_formset.urlpatterns
    + cv_question.urlpatterns
)



import django_tables2 as tables
from crispy_forms.layout import Row, Fieldset
from django.utils.translation import gettext as _

from app.models import Author, Detail
from crud_views.lib.crispy import Column4, CrispyModelForm, CrispyModelViewMixin, CrispyDeleteForm, Column8
from crud_views.lib.table import Table, UUIDLinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    MessageMixin,
    ListViewTableMixin,
    ListViewPermissionRequired,
    DeleteViewPermissionRequired,
)
from crud_views.lib.viewset import ViewSet

cv_detail = ViewSet(model=Detail, name="detail", icon_header="fa-solid fa-circle-info")


class DetailForm(CrispyModelForm):
    class Meta:
        model = Detail
        fields = "__all__"

    def get_layout_fields(self):
        return [
            Fieldset("Numbers", Row(Column4("integer"), Column4("number"))),
            Fieldset("Texts", Row(Column4("char"), Column8("text"))),
            Fieldset("Booleans", Row(Column4("boolean"), Column4("boolean_two"))),
            Fieldset("Date an time", Row(Column4("date"), Column4("date_time"))),
            Fieldset("References", Row(Column4("author"), Column4("foo"))),
        ]


class DetailTable(Table):
    id = UUIDLinkDetailColumn()
    integer = tables.Column()
    number = tables.Column()


class DetailListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Detail
    cv_viewset = cv_detail
    cv_context_actions = ListViewPermissionRequired.cv_context_actions
    table_class = DetailTable


class DetailCreateView(CrispyModelViewMixin, MessageMixin, CreateViewPermissionRequired):
    model = Author
    form_class = DetailForm
    cv_viewset = cv_detail
    cv_message = "Created detail »{object}«"


class DetailUpdateView(CrispyModelViewMixin, MessageMixin, UpdateViewPermissionRequired):
    model = Detail
    form_class = DetailForm
    cv_viewset = cv_detail
    cv_message = "Updated detail »{object}«"


class DetailDeleteView(CrispyModelViewMixin, MessageMixin, DeleteViewPermissionRequired):
    model = Detail
    form_class = CrispyDeleteForm
    cv_viewset = cv_detail
    cv_message = "Deleted detail »{object}«"


class DetailDetailView(DetailViewPermissionRequired):
    model = Detail
    cv_viewset = cv_detail

    cv_property_display = [
        {
            "title": _("Attributes"),
            "icon": "list",
            "description": _("All field types supported by the detail view"),
            "properties": [
                "integer",
                "number",
                {"path": "char", "detail": _("Short text field")},
                {"path": "text", "detail": _("Long text field")},
                "boolean",
                "boolean_two",
                "date",
                "date_time",
            ],
        },
        {
            "title": _("Extra"),
            "icon": "circle-info",
            "description": _("References, timestamps and computed values"),
            "properties": [
                {"path": "id", "detail": _("Unique identifier")},
                "author",
                "foo",
                {"path": "created_dt", "detail": _("When this record was created")},
                {"path": "modified_dt", "detail": _("Last modification timestamp")},
                {"path": "model_prop", "detail": _("Computed from a model property")},
            ],
        },
    ]



from __future__ import annotations

from crispy_forms.layout import LayoutObject
from django.template.loader import render_to_string


class Formsets(LayoutObject):
    template = "crud_views/formsets/formsets.html"

    def render(self, form, context, **kwargs):
        formsets = context.get("formsets")
        context.update(
            {
                "formsets": formsets,
            }
        )
        return render_to_string(self.template, context.flatten())


class FormControl(LayoutObject):
    template = "crud_views/formsets/control.html"

    def __init__(self, formset=None):
        self.formset = formset

    def render(self, form, context, **kwargs):
        context.update({"form": form, "formset": self.formset})
        return render_to_string(self.template, context.flatten())



from typing import Dict, Any

from pydantic import BaseModel, Field

from crud_views.lib.settings import crud_views_settings


def default_dict(*args, **kwargs) -> Any:
    return dict()


class ViewData(BaseModel):
    data: Dict[str, Any] = Field(default_factory=default_dict)  # cv_key to Any

    def set(self, key, value):
        self.data[key] = value

    def get(self, key):
        return self.data.get(key)


class ViewSetData(BaseModel):
    views: Dict[str, ViewData] = Field(default_factory=default_dict)  # ViewSet name to ViewData

    def get_view_data(self, view_key: str) -> ViewData:
        if view_key not in self.views.keys():
            view_data = ViewData()
            self.views[view_key] = view_data
            return view_data
        return self.views[view_key]


class SessionData(BaseModel):
    """
    Session data for django-viewset.
    Data ist stored in a hierarchy: app -> viewset -> view -> data
    """

    view: Any = None
    apps: Dict[str, ViewSetData] = Field(default_factory=default_dict)  # app ViewSetData

    @property
    def view_key(self) -> str:
        key = f"{self.view.cv_viewset.name}-{self.view.cv_key}"
        return key

    @property
    def app_label(self) -> str:
        return self.view.model._meta.app_label

    def get_viewset_data(self) -> ViewSetData:
        if self.app_label not in self.apps.keys():
            viewset_data = ViewSetData()
            self.apps[self.app_label] = viewset_data
            return viewset_data
        return self.apps[self.app_label]

    def get_view_data(self) -> ViewData:
        viewset = self.get_viewset_data()
        return viewset.get_view_data(self.view_key)

    @classmethod
    def from_view(cls, view):
        session_data = view.request.session.get(crud_views_settings.session_data_key, None)
        if session_data is None:
            session_data = SessionData(view=view)
        else:
            session_data = SessionData(view=view, **session_data)
        return session_data

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        serialized_data = self.model_dump(
            exclude={
                "view",
            }
        )
        self.view.request.session[crud_views_settings.session_data_key] = serialized_data

    def __setitem__(self, key: str, value: Any):
        data = self.get_view_data()
        data.data[key] = value

    def __getitem__(self, key: str) -> Any:
        data = self.get_view_data()
        return data.data[key]

    def __delitem__(self, key: str):
        data = self.get_view_data()
        del data.data[key]

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self[key]
        except KeyError:
            return default



from crud_views.lib.view import CrudViewPermissionRequiredMixin
from crud_views.lib.views.delete import DeleteView


class PolymorphicDeleteView(DeleteView):
    pass


class PolymorphicDeleteViewPermissionRequired(CrudViewPermissionRequiredMixin, PolymorphicDeleteView):
    cv_permission = "delete"



import pytest
from django.test.client import Client

from crud_views.lib.viewset import ViewSet


@pytest.mark.django_db
def test_manage_view_auto_registered(cv_author: ViewSet):
    """Manage view should be auto-registered when CRUD_VIEWS_MANAGE_VIEWS_ENABLED=yes."""
    assert cv_author.has_view("manage")


@pytest.mark.django_db
def test_manage_url_pattern(cv_author: ViewSet):
    """Manage URL should be in the viewset's urlpatterns."""
    routes = [p.name for p in cv_author.urlpatterns]
    assert "author-manage" in routes


@pytest.mark.django_db
def test_manage_view_accessible(client_user_author_view: Client, cv_author):
    """Manage view should be accessible (has_permission always returns True)."""
    response = client_user_author_view.get("/author/manage/")
    assert response.status_code == 200


@pytest.mark.django_db
def test_manage_view_shows_viewset_info(client_user_author_view: Client, cv_author):
    """Manage view should display viewset name and config."""
    response = client_user_author_view.get("/author/manage/")
    content = response.content.decode("utf-8")

    assert "author" in content
    assert "ViewSet" in content


@pytest.mark.django_db
def test_manage_view_shows_permissions(client_user_author_view: Client, cv_author):
    """Manage view should list the viewset's permissions."""
    response = client_user_author_view.get("/author/manage/")
    content = response.content.decode("utf-8")

    assert "add" in content
    assert "change" in content
    assert "delete" in content
    assert "view" in content


@pytest.mark.django_db
def test_manage_view_shows_registered_views(client_user_author_view: Client, cv_author):
    """Manage view should list all registered views for the viewset."""
    response = client_user_author_view.get("/author/manage/")
    content = response.content.decode("utf-8")

    for key in cv_author.get_all_views():
        assert key in content


@pytest.mark.django_db
def test_manage_context_has_cv(client_user_author_view: Client, cv_author):
    """Template context should contain the viewset as 'cv'."""
    response = client_user_author_view.get("/author/manage/")
    assert response.context["cv"] is cv_author


@pytest.mark.django_db
def test_manage_context_has_permission_rows(client_user_author_view: Client, cv_author):
    """Template context 'data' should contain permission rows with access info."""
    response = client_user_author_view.get("/author/manage/")
    data = response.context["data"]

    assert len(data) >= 4
    keys = [row["viewset"] for row in data]
    assert "add" in keys
    assert "view" in keys

    # user_author_view has view permission, so that row should show True
    view_row = next(r for r in data if r["viewset"] == "view")
    assert view_row["has_permission"] is True

    # user_author_view does NOT have add permission
    add_row = next(r for r in data if r["viewset"] == "add")
    assert add_row["has_permission"] is False


@pytest.mark.django_db
def test_manage_context_has_view_data(client_user_author_view: Client, cv_author):
    """Template context 'views' should contain view metadata dicts."""
    response = client_user_author_view.get("/author/manage/")
    views = response.context["views"]

    assert "list" in views
    assert "detail" in views
    assert "create" in views

    # Each view entry should have base, templates, icons sections
    list_view = views["list"]
    assert "base" in list_view
    assert "templates" in list_view
    assert "icons" in list_view
    assert list_view["base"]["cv_key"] == "list"


@pytest.mark.django_db
def test_manage_registered_on_all_viewsets(cv_author, cv_publisher, cv_book):
    """Manage view should be auto-registered on all viewsets in debug mode."""
    assert cv_author.has_view("manage")
    assert cv_publisher.has_view("manage")
    assert cv_book.has_view("manage")



# Generated by Django 5.2.11 on 2026-02-20 17:08

import django_fsm
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("app", "0005_alter_campaign_state"),
    ]

    operations = [
        migrations.AlterField(
            model_name="campaign",
            name="state",
            field=django_fsm.FSMField(
                choices=[
                    ("new", "New"),
                    ("active", "Active"),
                    ("success", "Success"),
                    ("canceled", "Cancelled"),
                    ("error", "Error"),
                ],
                default="new",
                max_length=50,
            ),
        ),
    ]



import django_tables2 as tables

from app.models import Baz
from crud_views.lib.table import Table, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
    ListViewPermissionRequired,
)
from crud_views.lib.viewset import ViewSet, ParentViewSet

cv_baz = ViewSet(model=Baz, name="baz", parent=ParentViewSet(name="bar"))


class BazTable(Table):
    id = LinkDetailColumn()
    name = tables.Column()


class BazListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Baz
    table_class = BazTable
    cv_viewset = cv_baz
    cv_list_actions = ["detail", "update", "delete"]


class BazDetailView(DetailViewPermissionRequired):
    model = Baz
    cv_viewset = cv_baz
    cv_property_display = [
        {
            "title": "Properties",
            "properties": ["id", "name"],
        },
    ]


class BazUpdateView(UpdateViewPermissionRequired):
    model = Baz
    fields = ["name"]
    cv_viewset = cv_baz


class BazCreateView(CreateViewPermissionRequired):
    model = Baz
    fields = ["name"]
    cv_viewset = cv_baz


class BazDeleteView(DeleteViewPermissionRequired):
    model = Baz
    cv_viewset = cv_baz



def user_viewset_permission(user, viewset, perm):
    from django.contrib.auth.models import Permission

    app, codename = viewset.permissions[perm].split(".")

    permission = Permission.objects.get(codename=codename)
    user.user_permissions.add(permission)
    user.save()



import pytest
from django.test.client import Client

from tests.test1.app.models import Publisher


# --- Create: validation errors ---


@pytest.mark.django_db
def test_create_missing_required_field(client_user_publisher_add: Client, cv_publisher):
    """POST with empty required field should re-render the form (200), not redirect."""
    response = client_user_publisher_add.post("/publisher/create/", {"name": ""})
    assert response.status_code == 200

    # Form should contain an error
    form = response.context["form"]
    assert form.errors
    assert "name" in form.errors

    # No object created
    assert Publisher.objects.count() == 0


@pytest.mark.django_db
def test_create_missing_field_entirely(client_user_publisher_add: Client, cv_publisher):
    """POST with missing field key should re-render the form with errors."""
    response = client_user_publisher_add.post("/publisher/create/", {})
    assert response.status_code == 200

    form = response.context["form"]
    assert form.errors
    assert "name" in form.errors
    assert Publisher.objects.count() == 0


# --- Update: validation errors ---


@pytest.mark.django_db
def test_update_blank_required_field(client_user_publisher_change: Client, cv_publisher, publisher_penguin):
    """POST update with empty required field should re-render the form."""
    pk = publisher_penguin.pk
    response = client_user_publisher_change.post(f"/publisher/{pk}/update/", {"name": ""})
    assert response.status_code == 200

    form = response.context["form"]
    assert form.errors
    assert "name" in form.errors

    # Original value unchanged
    publisher_penguin.refresh_from_db()
    assert publisher_penguin.name == "Penguin"


# --- Delete: confirm checkbox ---


@pytest.mark.django_db
def test_delete_without_confirm(client_user_publisher_delete: Client, cv_publisher, publisher_penguin):
    """POST delete without confirm checkbox should re-render the form."""
    pk = publisher_penguin.pk
    response = client_user_publisher_delete.post(f"/publisher/{pk}/delete/", {})
    assert response.status_code == 200

    form = response.context["form"]
    assert form.errors

    # Object not deleted
    assert Publisher.objects.filter(pk=pk).exists()


# --- CSRF ---


@pytest.mark.django_db
def test_csrf_enforced_on_create(cv_publisher, user_publisher_add):
    """POST without CSRF token should be rejected (403)."""
    client = Client(enforce_csrf_checks=True)
    client.force_login(user_publisher_add)

    response = client.post("/publisher/create/", {"name": "Test"})
    assert response.status_code == 403

    assert Publisher.objects.count() == 0


@pytest.mark.django_db
def test_csrf_enforced_on_update(cv_publisher, user_publisher_change, publisher_penguin):
    """POST update without CSRF token should be rejected (403)."""
    pk = publisher_penguin.pk
    client = Client(enforce_csrf_checks=True)
    client.force_login(user_publisher_change)

    response = client.post(f"/publisher/{pk}/update/", {"name": "Hacked"})
    assert response.status_code == 403

    publisher_penguin.refresh_from_db()
    assert publisher_penguin.name == "Penguin"


@pytest.mark.django_db
def test_csrf_enforced_on_delete(cv_publisher, user_publisher_delete, publisher_penguin):
    """POST delete without CSRF token should be rejected (403)."""
    pk = publisher_penguin.pk
    client = Client(enforce_csrf_checks=True)
    client.force_login(user_publisher_delete)

    response = client.post(f"/publisher/{pk}/delete/", {"confirm": True})
    assert response.status_code == 403

    assert Publisher.objects.filter(pk=pk).exists()


# --- Form re-render preserves input ---


@pytest.mark.django_db
def test_create_rerender_preserves_input(client_user_publisher_add: Client, cv_publisher):
    """On validation failure, the form should preserve the submitted values."""
    # name exceeds max_length (200) to trigger validation error
    long_name = "A" * 201
    response = client_user_publisher_add.post("/publisher/create/", {"name": long_name})
    assert response.status_code == 200

    form = response.context["form"]
    assert form.errors
    # The submitted value should be preserved in the form
    assert form.data["name"] == long_name



import django_tables2 as tables
from django.forms import modelform_factory

from app.models import Poly, PolyOne, PolyTwo
from crud_views_polymorphic.lib import (
    PolymorphicCreateSelectView,
    PolymorphicCreateViewPermissionRequired,
    PolymorphicUpdateViewPermissionRequired,
    PolymorphicDetailViewPermissionRequired,
)
from crud_views_polymorphic.lib.delete import PolymorphicDeleteViewPermissionRequired
from crud_views.lib.views import ListViewPermissionRequired
from crud_views.lib.table import Table
from crud_views.lib.views import ListViewTableMixin
from crud_views.lib.viewset import ViewSet

cv_poly = ViewSet(
    model=Poly,
    name="poly",
)


class PolyTable(Table):
    id = tables.Column()
    shared = tables.Column()


class PolyListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Poly
    table_class = PolyTable
    cv_viewset = cv_poly
    cv_list_actions = ["detail", "update", "delete"]
    cv_context_actions = [
        "create_select",
    ]


class PolyCreateSelectView(PolymorphicCreateSelectView):
    model = Poly
    # fields = ["shared"]
    cv_viewset = cv_poly


PolyOneFormCreate = modelform_factory(
    PolyOne,
    fields=[
        "shared",
        "one",
    ],
)

PolyTwoFormCreate = modelform_factory(
    PolyTwo,
    fields=[
        "shared",
        "two",
    ],
)


class PolyCreateView(PolymorphicCreateViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly
    polymorphic_forms = {PolyOne: PolyOneFormCreate, PolyTwo: PolyTwoFormCreate}


PolyOneFormUpdate = modelform_factory(
    PolyOne,
    fields=[
        "shared",
        "one",
    ],
)

PolyTwoFormUpdate = modelform_factory(
    PolyTwo,
    fields=[
        "shared",
        "two",
    ],
)


class PolyUpdateView(PolymorphicUpdateViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly
    polymorphic_forms = {PolyOne: PolyOneFormUpdate, PolyTwo: PolyTwoFormUpdate}


class PolyDeleteView(PolymorphicDeleteViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly


class PolyDetailView(PolymorphicDetailViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly
    cv_property_display = [
        {
            "title": "Properties",
            "properties": ["id", "shared"],
        },
    ]



from django.views import generic
from django.views.generic.detail import SingleObjectMixin

from crud_views.lib.view import CrudView
from crud_views.lib.viewset import path_regs, PrimaryKeys


class RedirectChildView(CrudView, SingleObjectMixin, generic.RedirectView):
    cv_key = "redirect_child"
    cv_path = "child"
    cv_backend_only = True

    # texts and labels
    cv_action_label_template: str | None = "crud_views/snippets/action/child.html"
    cv_action_short_label_template: str | None = "crud_views/snippets/action_short/child.html"

    def get_redirect_url(self, *args, **kwargs):
        obj = self.get_object()
        url = self.cv_get_child_url(self.cv_redirect, self.cv_redirect_key, obj)
        return url

    @classmethod
    def cv_get_url_extra_kwargs(cls) -> dict:
        return {"redirect": cls.cv_redirect, "redirect_key": cls.cv_redirect_key}

    @classmethod
    def cv_path_contribute(cls) -> str:
        """
        Contribute path to the path of the view
        """
        pr = path_regs.get_path("redirect", PrimaryKeys.KEY)
        pk = path_regs.get_path("redirect_key", PrimaryKeys.KEY)
        return f"{pr}/{pk}"



import pytest
from django.test.client import Client

from tests.lib.helper.boostrap5 import Table


@pytest.mark.django_db
def test_list_empty(client_user_author_view: Client, cv_author):
    response = client_user_author_view.get("/author/")
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 0


@pytest.mark.django_db
def test_list_multiple_authors(
    client_user_author_view: Client, cv_author, author_douglas_adams, author_terry_pratchett
):
    response = client_user_author_view.get("/author/")
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 2

    texts = [row.columns[1].text for row in table.rows]
    assert "Douglas" in texts
    assert "Terry" in texts


@pytest.mark.django_db
def test_list_table_headers(client_user_author_view: Client, cv_author, author_douglas_adams):
    response = client_user_author_view.get("/author/")
    assert response.status_code == 200

    table = Table(response)
    header_texts = [h.text for h in table.headers]

    assert "Id" in header_texts or "id" in [h.lower() for h in header_texts]
    assert "First name" in header_texts or "first_name" in [h.lower() for h in header_texts]
    assert "Last name" in header_texts or "last_name" in [h.lower() for h in header_texts]
    assert "Pseudonym" in header_texts or "pseudonym" in [h.lower() for h in header_texts]
    assert "Actions" in header_texts


@pytest.mark.django_db
def test_list_row_actions(client_user_author_view: Client, cv_author, author_douglas_adams):
    response = client_user_author_view.get("/author/")
    assert response.status_code == 200

    table = Table(response)
    row = table.rows[0]
    action_keys = [a.key for a in row.actions]

    assert "detail" in action_keys
    assert "update" in action_keys
    assert "delete" in action_keys



import django_tables2 as tables
from crispy_forms.layout import Row
from django.utils.translation import gettext_lazy as _

from app.models.poly import Poly, PolyOne, PolyTwo
from app.views.poly.one import PolyOneForm
from app.views.poly.two import PolyTwoForm
from crud_views.lib.crispy import Column4, CrispyModelViewMixin
from crud_views.lib.crispy.form import CrispyForm
from crud_views_polymorphic.lib import (
    PolymorphicCreateSelectView,
    PolymorphicCreateViewPermissionRequired,
    PolymorphicUpdateViewPermissionRequired,
    PolymorphicDetailViewPermissionRequired,
)
from crud_views_polymorphic.lib.create_select import PolymorphicContentTypeForm
from crud_views_polymorphic.lib.delete import PolymorphicDeleteViewPermissionRequired
from crud_views.lib.table import Table
from crud_views.lib.views import ListViewPermissionRequired
from crud_views.lib.views import ListViewTableMixin
from crud_views.lib.viewset import ViewSet

cv_poly = ViewSet(model=Poly, name="poly", icon_header="fa-solid fa-sun")


class PolyTable(Table):
    id = tables.Column()
    shared = tables.Column()


class PolyListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Poly
    table_class = PolyTable
    cv_viewset = cv_poly
    cv_list_actions = ["detail", "update", "delete"]
    cv_context_actions = [
        "create_select",
    ]


class CrispyPolymorphicContentTypeForm(CrispyForm, PolymorphicContentTypeForm):
    submit_label = _("Select")

    def get_layout_fields(self):
        return Row(Column4("polymorphic_ctype_id"))


class PolyCreateSelectView(CrispyModelViewMixin, PolymorphicCreateSelectView):
    model = Poly
    form_class = CrispyPolymorphicContentTypeForm
    cv_viewset = cv_poly


class PolyCreateView(CrispyModelViewMixin, PolymorphicCreateViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly
    cv_context_actions = ["home"]
    polymorphic_forms = {PolyOne: PolyOneForm, PolyTwo: PolyTwoForm}


class PolyUpdateView(CrispyModelViewMixin, PolymorphicUpdateViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly
    polymorphic_forms = {PolyOne: PolyOneForm, PolyTwo: PolyTwoForm}


class PolyDeleteView(PolymorphicDeleteViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly


class PolyDetailView(PolymorphicDetailViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly
    cv_property_display = [
        {
            "title": "Properties",
            "icon": "shapes",
            "description": "Shared polymorphic fields",
            "properties": ["id", "shared"],
        },
    ]



import uuid

from django.db import models
from django.utils.translation import gettext as _
from django_fsm import FSMField, transition
from ordered_model.models import OrderedModel
from polymorphic.models import PolymorphicModel

from crud_views_workflow.lib.enums import WorkflowComment
from crud_views_workflow.lib.enums import BadgeEnum
from crud_views_workflow.lib.mixins import WorkflowModelMixin


class Author(OrderedModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    first_name = models.CharField(max_length=100)
    last_name = models.CharField(max_length=100)
    pseudonym = models.CharField(max_length=100, blank=True, null=True)

    def __str__(self):
        return f"{self.first_name} {self.last_name}"

    @property
    def full_name(self):
        return f"{self.first_name} {self.last_name}"


class Publisher(models.Model):
    name = models.CharField(max_length=200)

    class Meta:
        ordering = ["name"]

    def __str__(self):
        return self.name


class Book(models.Model):
    title = models.CharField(max_length=200)
    publisher = models.ForeignKey(Publisher, on_delete=models.CASCADE, related_name="books")

    class Meta:
        ordering = ["title"]

    def __str__(self):
        return self.title


class CampaignState(models.TextChoices):
    NEW = "new", _("New")
    ACTIVE = "active", _("Active")
    SUCCESS = "success", _("Success")
    CANCELED = "canceled", _("Cancelled")
    ERROR = "error", _("Error")


class Campaign(WorkflowModelMixin, models.Model):
    STATE_CHOICES = CampaignState
    STATE_BADGES = {
        CampaignState.NEW: BadgeEnum.LIGHT,
        CampaignState.ACTIVE: BadgeEnum.INFO,
        CampaignState.SUCCESS: BadgeEnum.PRIMARY,
        CampaignState.CANCELED: BadgeEnum.WARNING,
        CampaignState.ERROR: BadgeEnum.DANGER,
    }

    name = models.CharField(max_length=128)
    state = FSMField(default=CampaignState.NEW, choices=CampaignState.choices)

    def __str__(self):
        return self.name

    @transition(
        field=state,
        source=CampaignState.NEW,
        target=CampaignState.ACTIVE,
        on_error=CampaignState.ERROR,
        custom={"label": "Activate", "comment": WorkflowComment.NONE},
    )
    def wf_activate(self, request=None, by=None, comment=None):
        pass

    @transition(
        field=state,
        source=CampaignState.ACTIVE,
        target=CampaignState.SUCCESS,
        on_error=CampaignState.ERROR,
        custom={"label": "Done", "comment": WorkflowComment.OPTIONAL},
    )
    def wf_done(self, request=None, by=None, comment=None):
        pass

    @transition(
        field=state,
        source=CampaignState.NEW,
        target=CampaignState.CANCELED,
        on_error=CampaignState.ERROR,
        custom={"label": "Cancel", "comment": WorkflowComment.REQUIRED},
    )
    def wf_cancel_new(self, request=None, by=None, comment=None):
        pass

    @transition(
        field=state,
        source=CampaignState.ACTIVE,
        target=CampaignState.CANCELED,
        on_error=CampaignState.ERROR,
        custom={"label": "Cancel", "comment": WorkflowComment.REQUIRED},
    )
    def wf_cancel_active(self, request=None, by=None, comment=None):
        pass


class Vehicle(PolymorphicModel):
    name = models.CharField(max_length=200)

    class Meta:
        ordering = ["name"]

    def __str__(self):
        return self.name


class Car(Vehicle):
    doors = models.IntegerField(default=4)


class Truck(Vehicle):
    payload_tons = models.IntegerField(default=10)



import django_tables2 as tables
from crispy_forms.layout import Row

from app.models import Bar
from crud_views.lib.crispy import CrispyModelForm, Column4, CrispyModelViewMixin, CrispyDeleteForm
from crud_views.lib.table import Table, LinkChildColumn, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
    ListViewPermissionRequired,
    CreateViewParentMixin,
)
from crud_views.lib.viewset import ViewSet, ParentViewSet

cv_bar = ViewSet(model=Bar, name="bar", parent=ParentViewSet(name="foo"), icon_header="fa-solid fa-bone")


class BarForm(CrispyModelForm):
    submit_label = "Create"

    class Meta:
        model = Bar
        fields = ["name"]

    def get_layout_fields(self):
        return Row(Column4("name"))


class BarTable(Table):
    id = LinkDetailColumn()
    name = tables.Column()
    baz = LinkChildColumn(name="baz", verbose_name="Baz", empty_values=())


class BarListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Bar
    table_class = BarTable
    cv_viewset = cv_bar
    cv_list_actions = ["detail", "update", "delete"]


class BarDetailView(DetailViewPermissionRequired):
    model = Bar
    cv_viewset = cv_bar
    cv_property_display = [
        {
            "title": "Properties",
            "icon": "bone",
            "description": "Bar attributes",
            "properties": [
                "id",
                "name",
            ],
        },
    ]


class BarUpdateView(CrispyModelViewMixin, UpdateViewPermissionRequired):
    model = Bar
    form_class = BarForm

    cv_viewset = cv_bar


class BarCreateView(CrispyModelViewMixin, CreateViewParentMixin, CreateViewPermissionRequired):
    model = Bar
    form_class = BarForm
    cv_viewset = cv_bar


class BarDeleteView(CrispyModelViewMixin, DeleteViewPermissionRequired):
    model = Bar
    form_class = CrispyDeleteForm
    cv_viewset = cv_bar



import django_tables2 as tables

from app.models import Bar
from crud_views.lib.table import Table, LinkChildColumn, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
)
from crud_views.lib.viewset import ViewSet, ParentViewSet

cv_bar = ViewSet(model=Bar, name="bar", parent=ParentViewSet(name="foo"))


class BarTable(Table):
    id = LinkDetailColumn()
    name = tables.Column()
    baz = LinkChildColumn(name="baz", verbose_name="Baz", empty_values=())


class BarListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Bar
    table_class = BarTable
    cv_viewset = cv_bar
    cv_list_actions = ["detail", "update", "delete"]


class BarDetailView(DetailViewPermissionRequired):
    model = Bar
    cv_viewset = cv_bar
    cv_property_display = [
        {
            "title": "Properties",
            "properties": ["id", "name"],
        },
    ]


class BarUpdateView(UpdateViewPermissionRequired):
    model = Bar
    fields = ["name"]
    cv_viewset = cv_bar


class BarCreateView(CreateViewPermissionRequired):
    model = Bar
    fields = ["name"]
    cv_viewset = cv_bar


class BarDeleteView(DeleteViewPermissionRequired):
    model = Bar
    cv_viewset = cv_bar
